import asyncio
import logging

from peacepie import msg_factory


class TickerAdmin:

    def __init__(self):
        self.tickers = {}
        self.ticker_index = 0

    def add_ticker(self, queue, period, delay, count, name, command):
        if not name:
            name = f'ticker_{self.ticker_index}'
            self.ticker_index += 1
        if not command:
            command = 'tick'
        self.tickers[name] = asyncio.get_running_loop().create_task(self.tick(queue, period, delay, count, command))
        return name

    def remove_ticker(self, name):
        task = self.tickers.get(name)
        if not task:
            return
        task.cancel()
        del self.tickers[name]

    def is_ticker_exists(self, name):
        return name in self.tickers.keys()

    async def tick(self, queue, period, delay, count, command):
        if delay:
            await asyncio.sleep(delay)
        while True:
            await queue.put(msg_factory.get_msg(command))
            await asyncio.sleep(period)
            if count:
                count -= 1
                if count == 0:
                    break



import asyncio

from peacepie import msg_factory


def start(timeout, queue, mid=None):
    asyncio.get_running_loop().create_task(wait(timeout, queue, mid))


async def wait(timeout, queue, mid):
    await asyncio.sleep(timeout)
    await queue.put(msg_factory.get_msg('timer', {'mid': mid}))





class AppStarter:

    def __init__(self):
        self.adaptor = None

    async def handle(self, msg):
        command = msg.get('command')
        if command == 'start':
            await self.start()
        else:
            return False
        return True

    async def start(self):
        name = 'initiator'
        body = {'class_desc': {'requires_dist': 'peacepie_example', 'class': 'Initiator'}, 'name': name}
        await self.adaptor.ask(self.adaptor.get_msg('create_actor', body), 120)
        body = {'params': [{'name': 'http_port', 'value': 9090}]}
        await self.adaptor.ask(self.adaptor.get_msg('set_params', body, name))
        await self.adaptor.send(self.adaptor.get_msg('start', None, name))



import binascii
import logging
import os

import rsa

from peacepie.assist import dir_opers

KEY_DIR = 'keys'


class RsaManager:

    def __init__(self):
        self.logger = logging.getLogger()
        self.pubkey = None
        self.prvkey = None
        dir_opers.makedir(KEY_DIR)
        if os.path.isfile(f'{KEY_DIR}/id_rsa.pub') and os.path.isfile(f'{KEY_DIR}/id_rsa.pub'):
            with open(f'{KEY_DIR}/id_rsa.pub', 'br') as f:
                data = f.read()
            self.pubkey = rsa.PublicKey.load_pkcs1(data)
            with open(f'{KEY_DIR}/id_rsa', 'br') as f:
                data = f.read()
            self.prvkey = rsa.PrivateKey.load_pkcs1(data)
        else:
            dir_opers.makedir(KEY_DIR, True)
            self.pubkey, self.prvkey = rsa.newkeys(256)
            with open(f'{KEY_DIR}/id_rsa.pub', 'bw') as f:
                f.write(rsa.PublicKey.save_pkcs1(self.pubkey))
            with open(f'{KEY_DIR}/id_rsa', 'bw') as f:
                f.write(rsa.PrivateKey.save_pkcs1(self.prvkey))

    def encode(self, value):
        res = None
        try:
            res = binascii.b2a_uu(rsa.encrypt(value.encode(), self.pubkey)).decode()
        except Exception as e:
            self.logger.exception(e)
        return res

    def decode(self, value):
        try:
            return rsa.decrypt(binascii.a2b_uu(value.encode()), self.prvkey).decode()
        except Exception as e:
            self.logger.exception(e)


instance = RsaManager()



import logging
import os

from peacepie.assist import log_util, dir_opers
from peacepie import multimanager

LOG_PATH = './logs/logs.logs'

instance = None


def init_loglistener(name):
    global instance
    instance = LogListener(name)


class LogListener:

    def __init__(self, name):
        self.logger = logging.getLogger()
        self.name = name
        queue = multimanager.instance.get_queue()
        if len(self.logger.handlers) == 0:
            add_default_handler(self.logger)
        handler = self.logger.handlers[0]
        self.log_desc = LogDesc(queue, handler.level)
        self.logger_listener = logging.handlers.QueueListener(queue, handler)
        self.logger_listener.start()
        self.is_working = True
        self.logger.info(f'{log_util.get_alias(self)} is created')

    def get_log_desc(self):
        return self.log_desc

    def exit(self):
        if not self.is_working:
            return
        self.is_working = False
        self.logger_listener.stop()
        self.logger.info(f'{log_util.get_alias(self)} is stopped')


def add_default_handler(logger):
    dir_opers.makedir(os.path.dirname(LOG_PATH))
    handler = logging.handlers.RotatingFileHandler(
        filename=LOG_PATH, mode='a', maxBytes=10485760, backupCount=5)
    formatter = logging.Formatter('%(levelname)-7s %(asctime)s %(processName)-11s %(lineno)4d %(module)-15s : %(message)s')
    handler.setFormatter(formatter)
    logger.setLevel('DEBUG')
    logger.addHandler(handler)


class LogDesc:

    def __init__(self, queue, level):
        self.queue = queue
        self.level = level

    def __repr__(self):
        res = f'{self.__class__.__name__}(queue={self.queue}, level={logging.getLevelName(self.level)})'
        return res



import shlex
import subprocess


def execute(cmd):
    if isinstance(cmd, str):
        cmd = ([cmd], 10)
    elif isinstance(cmd, list):
        cmd = (cmd, 10)
    timeout = cmd[1] if len(cmd) == 2 and isinstance(cmd[1], int) else 10
    processes = []
    for com in cmd[0]:
        args = shlex.split(com)
        stdin = None if len(processes) == 0 else processes[-1].stdout
        processes.append(subprocess.Popen(args, stdin=stdin, stdout=subprocess.PIPE, stderr=subprocess.PIPE))
    stdout, stderr = processes[-1].communicate(timeout=timeout)
    return_code = processes[-1].returncode
    stdout_str = stdout.decode('utf-8').strip()
    stderr_str = stderr.decode('utf-8').strip()
    return return_code, stdout_str, stderr_str


def exe(cmd):
    command, cwd, timeout = cmd
    result = subprocess.run(command, cwd=cwd, shell=True, capture_output=True, text=True, timeout=timeout)
    return_code = result.returncode
    stdout_str = result.stdout
    stderr_str = result.stderr
    return return_code, stdout_str, stderr_str





import asyncio
import json
import logging
import logging.config
import multiprocessing
import os
import sys
from logging.handlers import QueueHandler

from peacepie import msg_factory, loglistener, multimanager, params, adaptor
from peacepie.assist import log_util, misc
from peacepie.control import admin


class ProcessAdmin:

    def __init__(self, parent):
        self.logger = logging.getLogger()
        self.parent = parent
        self.processes = {}
        self.process_index = 0
        self.logger.info(log_util.get_alias(self) + ' is created')

    async def exit(self):
        await asyncio.gather(*[self.remove_process(complex_name) for complex_name in self.processes])

    def join_process(self, process):
        return process.join(timeout=4)

    async def remove_process(self, complex_name):
        process = self.processes.get(complex_name)
        node = complex_name.get_actor_name()
        if not process:
            logging.warning(f'The process for "{node}" is not found')
            return
        self.processes.pop(complex_name, None)
        await self.parent.adaptor.send(msg_factory.get_msg('remove_process', None, node))
        await asyncio.wait_for(
            asyncio.to_thread(self.join_process, process),
            timeout=5
        )
        if not process.is_alive():
            logging.info(f'The process for "{node}" is successfully completed')
            return
        process.terminate()
        try:
            await asyncio.to_thread(process.join, timeout=1)
            logging.info(f'The process for "{node}" is terminated')
        except asyncio.TimeoutError:
            process.kill()
            await asyncio.to_thread(process.join, timeout=1)
            logging.info(f'The process for "{node}" is killed')

    def get_members(self):
        res = [process.get_actor_name() for process in self.processes]
        res.append(self.parent.adaptor.name)
        res.sort()
        return res

    async def create_process(self, recipient):
        # name = misc.ComplexName(self.parent.host_name, f'process_{self.process_index}', 'admin')
        # self.process_index += 1
        name = misc.ComplexName(self.parent.host_name, None, 'admin')
        p = multiprocessing.Process(
            target=create,
            args=(self.parent.adaptor.name, name, params.instance,
                  msg_factory.instance.get_queue(), loglistener.instance.get_log_desc(), recipient))
        p.start()
        if name.process_name is None:
            name.process_name = p.name
        self.processes[name] = p


async def run_wrapper(actor):
    await actor.run()
    tasks = [t for t in asyncio.all_tasks() if not t.done()]
    for t in tasks:
        if t != asyncio.current_task():
            logging.warning(f'There is a pending task: {t}')


def create(lord, name, prms, msg_queue, log_desc, recipient):
    if name.process_name is None:
        name.process_name = multiprocessing.current_process().name
    params.instance = prms
    if params.instance.get('separate_log_per_process'):
        log_config()
    else:
        logger = logging.getLogger()
        while logger.hasHandlers():
            logger.removeHandler(logger.handlers[0])
        logger.addHandler(QueueHandler(log_desc.queue))
        logger.setLevel(log_desc.level)
    prefix = f'{name.host_name}.{name.process_name}'
    multimanager.init_multimanager(f'{prefix}.multimanager')
    msg_factory.init_msg_factory(name.host_name, name.process_name, 'msg_factory', msg_queue)
    performer = admin.Admin(lord, name.host_name, name.process_name, log_desc)
    try:
        actor = adaptor.Adaptor(None, name.get_actor_name(), None, performer, recipient)
        asyncio.run(run_wrapper(actor))
    except KeyboardInterrupt:
        pass
    except asyncio.CancelledError:
        pass
    except BaseException as ex:
        logging.exception(ex)


def log_config():
    config_filename = params.instance.get('log_config')
    process = ''
    if params.instance.get('separate_log_per_process'):
        process = f'/{multiprocessing.current_process().name}'
    try:
        with open(config_filename) as f:
            config = json.load(f)
        for _, handler_config in config.get('handlers', {}).items():
            if 'filename' in handler_config:
                handler_config['filename'] = f'{params.instance.get("log_dir")}{process}/{handler_config["filename"]}'
        check_paths(config)
        logging.config.dictConfig(config)
    except BaseException as ex:
        logging.exception(ex)


def check_paths(config):
    filenames = set([handler.get('filename') for handler in config.get('handlers').values()])
    filepaths = set([os.path.dirname(filename) for filename in filenames])
    for filepath in filepaths:
        if not os.path.exists(filepath):
            os.makedirs(filepath)
    if params.instance.get('developing_mode') or 'pycharm' in sys.executable.lower():
        for filename in filenames:
            try:
                os.remove(filename)
            except FileNotFoundError:
                pass



import multiprocessing
import os
import sys
import socket

from peacepie.assist import dir_opers, version
from peacepie.assist.auxiliaries import is_testing


instance = None
test_instance = None


def init_params(path, test_params):
    global instance
    global test_instance
    test_instance = test_params
    try:
        with open(path) as f:
            pass
    except (FileNotFoundError, TypeError):
        path = None
    if path is None:
        path = deploy_environment()
    params = []
    try:
        with open(path) as f:
            params = [line.strip().split('#')[0] for line in f.readlines()]
    except BaseException as bex:
        print(bex)
    res = {}
    for param in params:
        if params == '':
            continue
        lst = param.strip().split('=', 1)
        if len(lst) == 2 and not lst[0].strip() == '' and not lst[1].strip() == '':
            name = lst[0].strip()
            value = lst[1].strip()
            if name == 'developing_mode' or name == 'separate_log_per_process':
                value = value.lower() == 'true'
            elif name == 'inter_port':
                value = int(value)
            elif name == 'intra_port':
                value = int(value)
            elif name == 'package_dir':
                value = normalize(value)
            elif name == 'plugin_dir':
                value = normalize(value)
            elif name == 'log_dir':
                value = normalize(value)
            res[name] = value
    res['source_path'] = f'{res.get("package_dir")}/source'
    res['ip'] = get_ip()
    ver = version.version_from_string(f'{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}')
    res['python_version'] = ver
    if res.get('process_name') is None:
        res['process_name'] = multiprocessing.current_process().name
    instance = res


def normalize(value):
    return value[:-1] if value.endswith('/') else value


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        s.connect(('8.8.8.8', 1))
        res = s.getsockname()[0]
    except OSError:
        res = '127.0.0.1'
    finally:
        s.close()
    return res


def deploy_environment():
    if is_testing():
        return None
    src = f'{os.path.dirname(__file__)}/resources/config/'
    dst = f'{os.getcwd()}/config/'
    dir_opers.rem_dir(dst)
    dir_opers.copy_dir(src, dst)
    return dst + 'peacepie.cfg'



import logging

from peacepie import params
from peacepie.assist import log_util

mid_gen = 0

instance = None


def init_msg_factory(host_name, process_name, name, queue):
    global instance
    instance = MsgFactory(host_name, process_name, name, queue)


def get_msg(command, body=None, recipient=None, sender=None, timeout=None, group_mid=None):
    return instance.get_msg(command, body, recipient, sender, timeout, group_mid)


def get_group_mid():
    global mid_gen
    res = mid_gen
    mid_gen += 1
    return res


class MsgFactory:

    def __init__(self, host_name, process_name, name, queue):
        self.logger = logging.getLogger()
        self.name = f'{host_name}.{process_name}.{name}'
        self.system_name = params.instance['system_name']
        self.host_name = host_name
        self.process_name = process_name
        self.queue = queue
        self.logger.info(f'{log_util.get_alias(self)} is created')

    def get_queue(self):
        return self.queue

    def get_mid(self, group_mid):
        global mid_gen
        res = f'{self.system_name}.{self.host_name}.{self.process_name}.{mid_gen}'
        if group_mid:
            res += f'({group_mid})'
        mid_gen += 1
        return res

    def get_msg(self, command, body=None, recipient=None, sender=None, timeout=None, group_mid=None):
        res = {'mid': self.get_mid(group_mid), 'command': command, 'body': body, 'recipient': recipient,
               'sender': sender, 'timeout': timeout}
        return res



import asyncio
import logging

from peacepie import msg_factory, params
from peacepie.assist import log_util, serialization, timer
from peacepie.control.intra import intra_queue


class IntraLink:

    def __init__(self, parent):
        self.parent = parent
        self.host = params.instance.get('intra_host') if self.parent.is_head else params.instance.get('ip')
        self.port = params.instance.get('intra_port') if self.parent.is_head else 0
        self.server = None
        self.tasks = {}
        self.links = {}
        self.head = None
        logging.info(log_util.get_alias(self) + ' is created')

    async def exit(self):
        tasks = [asyncio.create_task(self.exiting(task)) for task in self.tasks]
        try:
            await asyncio.wait_for(asyncio.gather(*tasks), 0.8)
        except Exception as e:
            logging.exception(e)
        try:
            self.server.close()
            await asyncio.wait_for(self.server.wait_closed(), 0.4)
        except Exception as e:
            logging.exception(e)
        logging.info(f'{log_util.get_alias(self)} is stopped on port {self.port}')

    async def exiting(self, task):
        writer = self.tasks.get(task)
        if writer and not writer.is_closing():
            linklog = f' the channel {link_log(writer)} is closing in {task.get_name()}'
            logging.info(f'{log_util.get_alias(self)}: {linklog}')
            writer.close()
            try:
                await asyncio.wait_for(writer.wait_closed(), timeout=0.4)
            except Exception as e:
                logging.exception(e)
        if task.done():
            logging.info(f'{log_util.get_alias(self)}: the {task.get_name()} already done')
            return
        logging.info(f'{log_util.get_alias(self)}: the {task.get_name()} is cancelling')
        task.cancel()
        try:
            await asyncio.wait_for(task, timeout=0.4)
            logging.info(f'{log_util.get_alias(self)}: the {task.get_name()} was cancelled')
        except Exception as e:
            logging.exception(e)
        logging.info(f'{log_util.get_alias(self)}: the exit for {task.get_name()} completed')

    async def run(self, queue):
        await self.start_server(queue)
        if not self.parent.is_head:
            asyncio.get_running_loop().create_task(
                self.start_client(params.instance['intra_host'], params.instance['intra_port'], queue, True))

    async def start_server(self, queue):
        try:
            self.server = await asyncio.start_server(self.server_read, self.host, self.port)
            self.port = self.server.sockets[0].getsockname()[1]
            logging.info(f'{log_util.get_alias(self)} is started on port {self.port}')
            if self.parent.is_head:
                await queue.put(msg_factory.get_msg('ready'))
        except asyncio.CancelledError:
            pass
        except Exception as e:
            logging.exception(e)

    async def server_read(self, reader, writer):
        linklog = f' the channel {link_log(writer)} was created in {asyncio.current_task().get_name()}'
        logging.info(f'{log_util.get_alias(self)}: {linklog}')
        self.tasks[asyncio.current_task()] = writer
        serializer = serialization.Serializer()
        body = {'lord': self.parent.lord, 'name': self.parent.adaptor.name,
                'addr': {'host': self.host, 'port': self.port}}
        msg = msg_factory.get_msg('intra_link', body)
        writer.write(serializer.serialize(msg))
        await writer.drain()
        if msg.get('command') not in self.parent.adaptor.not_log_commands:
            logging.debug(log_util.sync_sent_log(self, msg))
        while reader:
            if reader.at_eof():
                break
            try:
                data = await reader.read(255)
                res = serializer.deserialize(data)
                if res:
                    await self.server_handle(res, writer)
            except Exception as ex:
                logging.exception(ex)
                reader = None
        linklog = f' the channel {link_log(writer)} was destroyed in {asyncio.current_task().get_name()}'
        del self.tasks[asyncio.current_task()]
        writer.close()
        await writer.wait_closed()
        logging.info(f'{log_util.get_alias(self)}: {linklog}')

    async def server_handle(self, msgs, writer):
        for msg in msgs:
            command = msg.get('command')
            if command not in self.parent.adaptor.not_log_commands:
                logging.debug(log_util.sync_received_log(self, msg))
            if command == 'intra_linked':
                await self.intra_linked(msg, writer)
            elif command == 'find_link':
                await self.find_link(msg, writer)
            else:
                recipient = self.clarify_recipient(msg.get('recipient'))
                if isinstance(recipient, asyncio.Queue):
                    await recipient.put(msg)
                    if command not in self.parent.adaptor.not_log_commands:
                        logging.debug(log_util.async_sent_log(self, msg))

    async def intra_linked(self, msg, writer):
        body = msg.get('body')
        name = None
        lord = None
        addr = None
        if body:
            name = body.get('name')
            lord = body.get('lord')
            addr = body.get('addr')
        self.links[name] = intra_queue.IntraQueue(lord, addr, writer)
        await self.parent.adaptor.notify(msg)

    async def find_link(self, msg, writer):
        link = self.links.get(msg['body']['name'])
        if link:
            ans = msg_factory.get_msg('link_found', {'addr': link.addr}, recipient=msg['sender'])
        else:
            ans = msg_factory.get_msg('link_not_found', recipient=msg['sender'])
        writer.write(serialization.Serializer.serialize(ans))
        await writer.drain()
        if ans.get('command') not in self.parent.adaptor.not_log_commands:
            logging.debug(log_util.sync_sent_log(self, ans))

    def clarify_recipient(self, recipient):
        if recipient is None:
            return self.parent.adaptor.queue
        if type(recipient) is dict:
            system_name = recipient.get('system')
            if system_name and system_name != self.parent.adaptor.get_param('system_name'):
                if self.parent.is_head:
                    return self.parent.interlink.queue
                else:
                    return self.links.get(self.head)
            else:
                recipient = recipient.get('entity')
                if not recipient:
                    return self.parent.adaptor.queue
        if type(recipient) is not str:
            return None
        if recipient.startswith('_'):
            res = self.parent.asks.get(recipient)
            return res
        else:
            if recipient == self.parent.adaptor.name:
                return self.parent.adaptor.queue
            else:
                return self.parent.actor_admin.get_actor_queue(recipient)

    async def start_client(self, host, port, queue, is_to_head):
        while True:
            try:
                reader, writer = await asyncio.open_connection(host, port)
                await self.client_read(reader, writer, queue, is_to_head)
            except asyncio.CancelledError:
                return
            except Exception as ex:
                logging.exception(ex)
            try:
               await asyncio.sleep(10)
            except asyncio.CancelledError:
                return

    async def client_read(self, reader, writer, queue, is_to_head):
        linklog = f' the channel {link_log(writer)} was created in {asyncio.current_task().get_name()}'
        logging.info(f'{log_util.get_alias(self)}: {linklog}')
        self.tasks[asyncio.current_task()] = writer
        serializer = serialization.Serializer()
        while reader:
            if reader.at_eof():
                break
            try:
                data = await reader.read(255)
                res = serializer.deserialize(data)
                if res:
                    await self.client_handle(res, writer, queue, is_to_head)
            except asyncio.CancelledError:
                pass
            except Exception as ex:
                logging.exception(ex)
                reader = None
        linklog = f' the channel {link_log(writer)} was destroyed in {asyncio.current_task().get_name()}'
        del self.tasks[asyncio.current_task()]
        writer.close()
        await writer.wait_closed()
        logging.info(f'{log_util.get_alias(self)}: {linklog}')

    async def client_handle(self, msgs, writer, queue, is_to_head=False):
        for msg in msgs:
            command = msg.get('command')
            if command not in self.parent.adaptor.not_log_commands:
                logging.debug(log_util.sync_received_log(self, msg))
            if command == 'intra_link':
                body = msg.get('body')
                name = None
                addr = None
                if body:
                    name = body.get('name')
                    addr = body.get('addr')
                if is_to_head:
                    self.head = name
                self.links[name] = intra_queue.IntraQueue(self.parent.lord, addr, writer)
                body = {'lord': self.parent.lord, 'name': self.parent.adaptor.name,
                        'addr': {'host': self.host, 'port': self.port}}
                ans = msg_factory.get_msg('intra_linked', body)
                await self.links[name].put(ans)
                if ans.get('command') not in self.parent.adaptor.not_log_commands:
                    logging.debug(log_util.sync_sent_log(self, ans))
                await queue.put(msg_factory.get_msg('ready'))
            elif command == 'intra_linked':
                body = msg.get('body')
                name = None
                lord = None
                addr = None
                if body:
                    name = body.get('name')
                    lord = body.get('lord')
                    addr = body.get('addr')
                self.links[name] = intra_queue.IntraQueue(lord, addr, writer)
                await self.parent.adaptor.notify(msg)
            else:
                recipient = self.clarify_recipient(msg.get('recipient'))
                if isinstance(recipient, asyncio.Queue):
                    await recipient.put(msg)
                    if msg.get('command') not in self.parent.adaptor.not_log_commands:
                        logging.debug(log_util.async_sent_log(self, msg))

    async def get_intra_queue(self, name):
        res = self.links.get(name)
        if res:
            return res
        if self.head is None:
            return None
        msg = msg_factory.get_msg('find_link', {'name': name})
        ans = await self.ask(msg)
        if ans['command'] != 'link_found':
            return None
        queue = asyncio.Queue()
        asyncio.get_running_loop().create_task(
            self.start_client(ans['body']['addr']['host'], ans['body']['addr']['port'], queue, False))
        await queue.get()
        return self.links.get(name)

    def get_recipients(self):
        res = [link for link in self.links.values()]
        return res

    async def ask(self, msg):
        entity = f'_{self.parent.ask_index}'
        self.parent.ask_index += 1
        msg['sender'] = {'node': self.parent.adaptor.name, 'entity': entity}
        queue = asyncio.Queue()
        self.parent.asks[entity] = queue
        await self.links[self.head].put(msg)
        if msg.get('command') not in self.parent.adaptor.not_log_commands:
            logging.debug(log_util.sync_ask_log(self, msg))
        timer.start(2, queue, msg['mid'])
        ans = await queue.get()
        command = ans.get('command')
        if command == 'timeout':
            logging.warning(log_util.async_received_log(self, ans))
        elif command not in self.parent.adaptor.not_log_commands:
            logging.debug(log_util.sync_received_log(self, ans))
        if entity is not None:
            del self.parent.asks[entity]
        return ans

    def get_members(self):
        res = [link[0] for link in self.links.items() if not link[1].lord]
        if not self.parent.lord:
            res.append(self.parent.adaptor.name)
        res.sort()
        return res

    def get_all_nodes(self):
        res = [key for key in self.links]
        res.append(self.parent.adaptor.name)
        return res


def link_log(writer):
    return f'{writer.get_extra_info("sockname")}<=>{writer.get_extra_info("peername")}'



import asyncio
import logging
import time

from peacepie import msg_factory, params
from peacepie.assist import log_util, serialization
from peacepie.control.inter import inter_queue


class InterServer:

    def __init__(self, parent):
        self.parent = parent
        self.system_name = params.instance['system_name']
        self.server = None
        self.host = None
        self.port = None
        self.links = {}
        self.queue = asyncio.Queue()
        logging.info(log_util.get_alias(self) + ' is created')

    async def run(self, queue):
        self.host = params.instance.get('ip')
        self.port = params.instance.get('inter_port')
        await self.start_server(queue)
        while True:
            msg = await self.queue.get()
            logging.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    logging.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                logging.exception(ex)

    async def handle(self, msg):
        command = msg.get('command')
        if command == 'inter_connect':
            await self.inter_connect(msg)
        elif command == 'inter_disconnect':
            await self.inter_disconnect(msg)
        else:
            await self.route(msg)
        return True

    async def start_server(self, queue):
        try:
            self.server = await asyncio.start_server(self.server_handle, self.host, self.port)
            await queue.put(msg_factory.get_msg('ready'))
            logging.info(f'{log_util.get_alias(self)} is started at address ({self.host}:{self.port})')
        except Exception as ex:
            logging.exception(ex)

    async def exit(self):
        self.server.close()
        await self.server.wait_closed()
        logging.info(f'InterServer on port {self.port} is closed')

    async def server_handle(self, reader, writer):
        peer = writer.get_extra_info('socket').getpeername()
        logging.warning(log_util.get_alias(self) + f' Channel to ({peer[0]}, {peer[1]}) is opened')
        serializer = serialization.Serializer()
        body = self.parent.adaptor.get_addr(self.system_name, self.parent.adaptor.name, None)
        msg = msg_factory.get_msg('inter_link', body)
        writer.write(serializer.serialize(msg))
        await writer.drain()
        logging.debug(log_util.sync_sent_log(self, msg))
        while reader:
            if reader.at_eof():
                break
            try:
                data = await reader.read(255)
                res = serializer.deserialize(data)
                if res:
                    await self._handle(res, writer)
            except Exception as ex:
                logging.exception(ex)
                reader = None
        writer.close()
        await writer.wait_closed()
        logging.warning(log_util.get_alias(self) + f' Channel to ({peer[0]}, {peer[1]}) is closed')

    async def inter_connect(self, msg):
        body = msg.get('body')
        addr = body.get('addr')
        host = addr.get('host')
        port = int(addr.get('port'))
        queue = asyncio.Queue()
        asyncio.get_running_loop().create_task(self.start_client(host, port, queue))
        ans = await queue.get()
        ans['recipient'] = msg.get('sender')
        await self.parent.adaptor.send(ans)

    async def inter_disconnect(self, msg):
        body = msg.get('body')
        if not body:
            return
        system_name = body.get('system')
        if not system_name:
            return
        link = self.links.get(system_name)
        if not link:
            return
        await link.close()
        del self.links[system_name]
        ans = self.parent.adaptor.get_msg('inter_disconnected', recipient=msg.get('sender'))
        await self.parent.adaptor.send(ans)
        logging.info(log_util.get_alias(self) + f' Channel to system "{system_name}" is closed')

    async def close_writer(self, writer):
        if not writer:
            return
        peer = writer.get_extra_info('socket').getpeername()
        writer.close()
        await writer.wait_closed()
        logging.warning(log_util.get_alias(self) + f' Channel to ({peer[0]}, {peer[1]}) is closed')

    async def start_client(self, host, port, queue):
        t = time.time()
        writer = None
        while True:
            try:
                reader, writer = await asyncio.open_connection(host, port)
                logging.info(log_util.get_alias(self) + f' Channel to ({host}, {port}) is opened')
                await self.client_handle(reader, writer, queue)
            except Exception as ex:
                logging.exception(ex)
            await self.close_writer(writer)
            await asyncio.sleep(1 if time.time() - t < 2 else 60)

    async def client_handle(self, reader, writer, queue):
        serializer = serialization.Serializer()
        while reader:
            if reader.at_eof():
                break
            try:
                data = await reader.read(255)
                res = serializer.deserialize(data)
                if res:
                    await self._handle(res, writer, queue)
            except Exception as ex:
                logging.exception(ex)
                reader = None

    async def _handle(self, msg, writer, queue=None):
        logging.debug(log_util.sync_received_log(self, msg))
        command = msg['command']
        if command == 'inter_link':
            await self.inter_link(msg, writer, queue)
        elif command == 'inter_linked':
            await self.inter_linked(msg, writer, queue)
        else:
            recipient = await self.clarify_recipient(msg['recipient'])
            if not recipient:
                return
            await recipient.put(msg)
            if isinstance(recipient, asyncio.Queue):
                logging.debug(log_util.async_sent_log(self, msg))
            else:
                logging.debug(log_util.sync_sent_log(self, msg))
        return True

    async def inter_link(self, msg, writer, queue):
        res = inter_queue.InterQueue(writer)
        system_addr = msg.get('body')
        system_name = system_addr.get('system')
        self.links[system_name] = res
        body = self.parent.adaptor.get_addr(self.system_name, self.parent.adaptor.name, None)
        ans = msg_factory.get_msg('inter_linked', body)
        await res.put(ans)
        logging.debug(log_util.sync_sent_log(self, ans))
        msg['command'] = 'inter_linked'
        await self.parent.adaptor.notify(msg)
        if queue:
            await queue.put(msg_factory.get_msg('inter_connected', system_addr))

    async def inter_linked(self, msg, writer, queue):
        res = inter_queue.InterQueue(writer)
        system_addr = msg.get('body')
        system_name = system_addr.get('system')
        self.links[system_name] = res
        await self.parent.adaptor.notify(msg)
        if queue:
            await queue.put(msg_factory.get_msg('inter_connected', system_addr))

    async def route(self, msg):
        res = self.links.get(msg['recipient'].get('system'))
        if msg.get('sender'):
            if res:
                msg['sender'] = self.parent.adaptor.add_system(msg['sender'], self.system_name)
            else:
                ans = msg_factory.get_msg('system_is_not_registered', recipient=msg['sender'])
                await self.parent.adaptor.send(ans, self)
                return
        await res.put(msg)
        logging.debug(log_util.sync_sent_log(self, msg))

    async def clarify_recipient(self, recipient):
        if type(recipient) is dict:
            node = recipient.get('node')
            if node and node != self.parent.adaptor.name:
                res = await self.parent.intralink.get_intra_queue(node)
                return res
            recipient = recipient.get('entity')
        if not recipient:
            return self.parent.adaptor.queue
        if type(recipient) is not str:
            return None
        if recipient.startswith('_'):
            res = self.parent.asks[recipient]
            return res
        else:
            if recipient == self.parent.adaptor.name:
                return self.parent.adaptor.queue
            else:
                return self.parent.actor_admin.get_actor_queue(recipient)



import enum

from peacepie.assist import json_util


class Serializer:

    class State(enum.Enum):
        MARKER = enum.auto()
        BODY = enum.auto()

    def __init__(self):
        self.state = self.State.MARKER
        self.length = 0
        self.data = b''

    def clear(self):
        self.state = self.State.MARKER
        self.length = 0
        self.data = b''

    @staticmethod
    def serialize(data):
        buf = json_util.json_dumps(data).encode('utf-8')
        return b'\xff\xfe' + (len(buf)).to_bytes(2, byteorder='big') + b'\x7f\xff' + buf

    def deserialize(self, data):
        self.data += data
        messages = []
        while True:
            if self.state is self.State.MARKER:
                if len(self.data) < 6:
                    break
                pos = self.data.find(b'\xff\xfe')
                if pos == -1:
                    break
                self.data = self.data[pos + 2:]
                if self.data[2:4] != b'\x7f\xff':
                    break
                self.length = int.from_bytes(self.data[:2], byteorder='big')
                self.data = self.data[4:]
                self.state = self.State.BODY
            if self.state is self.State.BODY:
                if len(self.data) < self.length:
                    break
                buf = self.data[:self.length]
                self.data = self.data[self.length:]
                self.state = self.State.MARKER
                self.length = 0
                res = json_util.json_loads(buf)
                messages.append(res)
        return messages if messages else None




class SeriesAdmin:

    def __init__(self):
        self.series = {}

    def next(self, name):
        series = self.series.get(name)
        if not series:
            series = {'value': 0, 'max_value': 0}
            self.series[name] = series
        value = series.get('value')
        max_value = series.get('max_value')
        if max_value > 0:
            if value > max_value:
                value = 0
                series['value'] = value
        series['value'] = value + 1
        return value

    def set_value(self, name, value):
        series = self.series.get(name)
        if not series:
            self.series[name] = {'value': value, 'max_value': 0}
        else:
            series['value'] = value

    def set_max_value(self, name, max_value):
        series = self.series.get(name)
        if not series:
            self.series[name] = {'value': 0, 'max_value': max_value}
        else:
            series['max_value'] = max_value


instance = SeriesAdmin()



import asyncio
import logging
import multiprocessing

from peacepie import adaptor, msg_factory
from peacepie.assist import log_util, timer

index = 0
replica_index = 0


class ActorLoader:

    def __init__(self, parent):
        global index
        self.name = f'actor_loader_{index}'
        index += 1
        self.parent = parent
        self.grandparent = parent.parent
        self.queue = asyncio.Queue()
        self.not_log_commands = set()
        self.cumulative_commands = {}
        logging.info(log_util.get_alias(self) + ' is created')

    async def run(self):
        while True:
            msg = await self.queue.get()
            logging.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    logging.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                logging.exception(ex)

    async def handle(self, msg):
        command = msg.get('command')
        if command == 'get_class':
            await self.get_class(msg)
        elif command == 'create_actor':
            await self.create_actor(msg, False)
        elif command == 'create_replica':
            await self.create_actor(msg, True)
        elif command == 'create_actors':
            await self.create_actors(msg)
        else:
            return False
        return True

    async def get_class(self, msg):
        clss = await self._get_class(msg)
        ans = msg_factory.get_msg('class', clss, recipient=msg.get('sender'))
        await self.grandparent.adaptor.send(ans, self)

    async def create_actor(self, msg, is_replica):
        global replica_index
        body = msg.get('body') if msg.get('body') else {}
        class_desc = body.get('class_desc')
        name = body.get('name')
        if self.parent.actors.get(name) and not is_replica:
            answer = msg_factory.get_msg('actor_is_not_created', recipient=msg.get('sender'))
            await self.grandparent.adaptor.send(answer, self)
            return
        real_name = name
        if is_replica and name is None:
            real_name = (f'replica_{self.grandparent.adaptor.get_param("host_name")}_'
                         f'{multiprocessing.current_process().name}_{replica_index}')
            replica_index += 1
        try:
            clss = await self._get_class(msg)
            adptr = adaptor.Adaptor(class_desc, real_name, self.grandparent, clss(), msg.get('sender'))
            if is_replica:
                adptr.pause_event = asyncio.Event()
        except Exception as e:
            logging.exception(e)
            answer = msg_factory.get_msg('actor_is_not_created', recipient=msg.get('sender'))
            await self.grandparent.adaptor.send(answer, self)
            return
        task = asyncio.get_running_loop().create_task(adptr.run())
        self.parent.actors[real_name] = {'adaptor': adptr, 'task': task}

    async def create_actors(self, msg):
        clss = await self._get_class(msg)
        queue = asyncio.Queue()
        actors = {}
        body = msg.get('body') if msg.get('body') else {}
        class_desc = body.get('class_desc')
        if not body:
            return
        names = body.get('names')
        if not names:
            return
        for name in names:
            try:
                adptr = adaptor.Adaptor(class_desc, name, self.grandparent, clss(), queue)
                task = asyncio.get_running_loop().create_task(adptr.run())
                actors[name] = {'adaptor': adptr, 'task': task}
            except Exception as e:
                logging.exception(e)
                await self.clear(actors, msg.get('sender'))
                return
        timeout = msg.get('timeout')
        if not timeout:
            timeout = 1
        timer.start(timeout, queue, msg.get('mid'))
        count = 0
        while True:
            if count == len(actors):
                break
            ans = await queue.get()
            logging.debug(log_util.async_received_log(self, ans))
            if ans.get('command') != 'actor_is_created':
                break
            count += 1
        if count < len(actors):
            await self.clear(actors, msg.get('sender'))
            return
        self.parent.actors.update(actors)
        system = self.grandparent.adaptor.get_param('system_name')
        node = self.grandparent.adaptor.name
        body = {'system': system, 'node': node, 'names': names}
        ans = msg_factory.get_msg('actors_are_created', body, msg.get('sender'))
        await self.grandparent.adaptor.send(ans, self)

    async def clear(self, actors, recipient):
        for actor in actors:
            actor['task'].cancel()
        ans = msg_factory.get_msg('actors_are_not_created', recipient=recipient)
        await self.grandparent.adaptor.send(ans, self)

    async def _get_class(self, msg):
        body = msg.get('body')
        if not body:
            return None
        class_desc = body.get('class_desc')
        if not class_desc:
            return None
        if isinstance(class_desc, type):
            return class_desc
        return await self.parent.package_admin.get_class(class_desc, msg.get('timeout'))



import logging
import re
import html


VERSION = 'version'
MAJOR_LEVEL = 'major'
MINOR_LEVEL = 'minor'
MICRO_LEVEL = 'micro'
VERSION_LEVELS = (MAJOR_LEVEL, MINOR_LEVEL, MICRO_LEVEL)


def version_to_string(val):
    return f'{val[MAJOR_LEVEL]}.{val[MINOR_LEVEL]}.{val[MICRO_LEVEL]}'


def version_from_string(val):
    if not val:
        return None
    is_prerelease = False
    suffix_start = 0
    for i, char in enumerate(val):
        if not char.isdigit() and char != '.':
            suffix_start = i
            is_prerelease = True
            break
    base_version = val[:suffix_start] if is_prerelease else val
    parts = base_version.split('.')
    major = int(parts[0]) if len(parts) > 0 else 0
    minor = int(parts[1]) if len(parts) > 1 else 0
    micro = int(parts[2]) if len(parts) > 2 else 0
    return {MAJOR_LEVEL: major, MINOR_LEVEL: minor, MICRO_LEVEL: micro}


def conditions_to_string(conditions):
    res = ''
    if conditions:
        res = ', '.join(f'{key}{version_to_string(value)}' for key, value in conditions.items())
    return res


def conditions_from_string(val):
    constraints = val.split(',')
    parsed = []
    for constraint in constraints:
        constraint = constraint.strip()
        if not constraint:
            continue
        for op in ['!=', '>=', '<=', '>', '<', '==', '~=', '^=']:
            if constraint.startswith(op):
                ver = constraint[len(op):]
                parsed.append({op: version_from_string(ver)})
                break
        else:
            parsed.append({'==': constraint})
    return parsed


def version_key(v):
    return tuple(map(int, v.split('.')))


def find_max_version(vers, conditions):
    if isinstance(conditions, str):
        conditions = conditions_from_string(conditions)
    res = None
    for val in vers:
        ver = version_from_string(val)
        if check_version(ver, conditions):
            if res is None:
                res = ver, val
            elif check_version(ver, {'>': res}):
                res = ver, val
    if res is None:
        return None
    else:
        return res[1]


def check_version(ver, conditions):
    try:
        res = _check_version(ver, conditions)
        return res
    except Exception as ex:
        logging.info(f'ver: {ver}, conditions: {conditions}')
        logging.exception(ex)
        return False


def _check_version(version, conditions):
    if isinstance(version, str):
        version = version_from_string(version)
    if isinstance(conditions, str):
        if conditions.startswith('(') and conditions.endswith(')'):
            conditions = conditions[1:-1]
        conditions = conditions_from_string(conditions)
    if not version:
        return True
    if not conditions:
        return True
    for condition in conditions:
        key, value = next(iter(condition.items()))
        if key == '==':
            for level in VERSION_LEVELS:
                if version[level] != value[level]:
                    return False
        elif key == '>':
            for level in VERSION_LEVELS:
                if version[level] > value[level]:
                    break
                elif version[level] < value[level]:
                    return False
        elif key == '<':
            for level in VERSION_LEVELS:
                if version[level] < value[level]:
                    break
                elif version[level] > value[level]:
                    return False
        elif key == '>=':
            for level in VERSION_LEVELS:
                if version[level] > value[level]:
                    break
                elif version[level] < value[level]:
                    return False
        elif key == '<=':
            for level in VERSION_LEVELS:
                if version[level] < value[level]:
                    break
                elif version[level] > value[level]:
                    return False
        elif key == '!=':
            res = False
            for level in VERSION_LEVELS:
                if version[level] != value[level]:
                    res = True
                    break
            return res
        elif key == '~=':
            for level in (MAJOR_LEVEL, MINOR_LEVEL):
                if version[level] != value[level]:
                    return False
        else:
            return False
    return True


pattern = r"""
    ^
    (?P<package_name>[a-zA-Z0-9_-]+(?:\.[a-zA-Z0-9_-]+)*)
    (?P<version_spec>[^;]*)
    (?:;\s*(?P<markers>.+))?
    $
"""

python_version_pattern = r"""
    python_version\s*
    (?P<constraints>(?:(?:<|<=|==|>=|>|!=)\s*['"]?[0-9._*]+['"]?(?:\s*,\s*)?)+)
"""

platform_system_pattern = r"""
    platform_system\s*
    ==\s*
    ['"](?P<platform_name>[a-zA-Z0-9_-]+)['"]
"""

extra_pattern = r"""
    extra\s*
    ==\s*
    ['"](?P<extra_name>[a-zA-Z0-9_-]+)['"]
"""

regex = re.compile(pattern, re.VERBOSE | re.IGNORECASE)
python_version_regex = re.compile(python_version_pattern, re.VERBOSE | re.IGNORECASE)
platform_system_regex = re.compile(platform_system_pattern, re.VERBOSE | re.IGNORECASE)
extra_regex = re.compile(extra_pattern, re.VERBOSE | re.IGNORECASE)


def parse_requires_dist(requires_dist):
    decoded_str = html.unescape(requires_dist)
    match = regex.match(decoded_str.strip())
    if match:
        result = match.groupdict()
        result = {k: v.strip() if v else None for k, v in result.items()}
        markers = result.pop('markers', None)
        if markers:
            py_matches = python_version_regex.finditer(markers)
            constraints = []
            for match in py_matches:
                for constraint in re.split(r'\s*,\s*', match.group('constraints')):
                    constraint = re.sub(r"\s*['\"]", "", constraint.strip())
                    constraint = re.sub(r"\s+", "", constraint)
                    if constraint:
                        constraints.append(constraint)
            if constraints:
                result['python_version'] = ",".join(constraints)
            platform_match = platform_system_regex.search(markers)
            if platform_match:
                result['platform_system'] = platform_match.group('platform_name')
            extra_match = extra_regex.search(markers)
            if extra_match:
                result['extra'] = extra_match.group('extra_name')
        return result
    else:
        return None


def to_canonical(name: str) -> str:
    name = name.lower()
    name = re.sub(r'[-_.]+', '-', name)
    name = re.sub(r'[^a-z0-9-]', '', name)
    name = name.strip('-')
    return name



from peacepie.assist.serialization import Serializer


class IntraQueue:

    def __init__(self, lord, addr, writer):
        self.lord = lord
        self.addr = addr
        self.writer = writer

    async def put(self, msg):
        self.writer.write(Serializer.serialize(msg))
        await self.writer.drain()

    async def close(self):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()
        self.writer = None

    def __repr__(self):
        res = f'{self.__class__.__name__}(lord={self.lord}, addr={self.addr}, '
        res += f'writer={self.writer.__class__.__name__}({id(self.writer)}))'
        return res




class ComplexName:

    @staticmethod
    def parse_name(name):
        tokens = name.split('.')
        if len(tokens) != 3:
            return None
        res = ComplexName(tokens[0], tokens[1], tokens[2])
        return res

    def __init__(self, host_name, process_name, actor_name):
        self.host_name = host_name
        self.process_name = process_name
        self.actor_name = actor_name

    def get_process_name(self):
        return f'{self.host_name}.{self.process_name}'

    def get_actor_name(self):
        return f'{self.host_name}.{self.process_name}.{self.actor_name}'

    def __repr__(self):
        return f'{self.host_name}:{self.process_name}:{self.actor_name}'

    def __eq__(self, other):
        if isinstance(other, ComplexName):
            return (self.host_name == other.host_name and
                    self.process_name == other.process_name and
                    self.actor_name == other.actor_name)
        return False

    def __hash__(self):
        return hash((self.host_name, self.process_name, self.actor_name))


class InterAddress:

    def __init__(self, host, port, authkey):
        self.host = host
        self.port = port
        self.authkey = authkey



import os
import re

from peacepie import msg_factory, params
from peacepie.assist import dir_opers


class InternalStarter:

    def __init__(self):
        self.adaptor = None

    async def handle(self, msg):
        command = msg.get('command')
        if command == 'start':
            await self.start()
        elif command == 'app_starter':
            await self.app_starter(msg)
        else:
            return False
        return True

    async def start(self):
        src = os.path.abspath(params.instance.get('starter'))
        package_name = os.path.basename(src)
        ans = await self.adaptor.ask(self.adaptor.get_msg('get_work_path'))
        body = ans.get('body')
        if not body:
            return
        dst = os.path.abspath(f'{body.get("path")}/{package_name}')
        os.symlink(src, dst)
        package_name = package_name.split('.')[0]
        msg = msg_factory.get_msg('create_actor', {'class_desc': {'requires_dist': package_name}, 'name': 'starter'})
        ans = await self.adaptor.ask(msg)
        command = ans.get('command')
        if command != 'actor_is_created':
            return
        body = {'internal_starter': self.adaptor.get_self_addr()}
        msg = msg_factory.get_msg('start', body, recipient=ans.get('body'))
        await self.adaptor.send(msg)
        await self.adaptor.send(self.adaptor.get_msg('remove_actor', {'name': self.adaptor.name}))

    async def app_starter(self, msg):
        body = msg.get('body')
        name = body.get('name')
        pattern = rf'async\s+def\s+{name}\(self,\s*msg\s*\):'
        with open(params.instance.get('starter'), 'r') as file:
            lines = file.readlines()
        beg = None
        end = len(lines)
        for i, line in enumerate(lines):
            if beg:
                if re.search(pattern, line):
                    end = i
                    break
            else:
                if re.search(pattern, line):
                    beg = i
                    pattern = r'def\s+\w+\(.*\):'
        if beg:
            lines[beg+1:end] = change_leading_spaces(lines[beg], body.get('txt'))
            with open(params.instance.get('starter'), 'w') as file:
                file.writelines(lines)
            await self.adaptor.send(self.adaptor.get_msg('app_starter_is_ready', recipient=msg.get('sender')))
        else:
            await self.adaptor.send(self.adaptor.get_msg('app_starter_is_not_ready', recipient=msg.get('sender')))


def change_leading_spaces(anchor, txt):
    res = txt.split('\n')
    res = [line + '\n' for line in res if line.strip()]
    dx = len(anchor) - len(anchor.lstrip()) + 4
    dx -= len(res[0]) - len(res[0].lstrip())
    res = [' ' * (len(line) - len(line.lstrip()) + dx) + line.lstrip() for line in res]
    res.append('\n')
    return res



import asyncio
import logging
import os
import platform
import shutil
import sys
from pathlib import Path

from peacepie.assist import json_util, version


def recreatedir(dirpath):
    try:
        if os.path.exists(dirpath):
            shutil.rmtree(dirpath)
            logging.info(f'Old folder "{dirpath}" is deleted')
        os.makedirs(dirpath)
        logging.info(f'Folder "{dirpath}" is created')
        logging.info(f'Folder "{dirpath}" is {"" if len(os.listdir(dirpath)) == 0 else "not "}empty')
    except Exception as e:
        logging.exception(e)


def makedir(dirpath, clear=False):
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)
        logging.info(f'Folder "{dirpath}" is created')
    elif clear:
        cleardir(dirpath)
        logging.info(f'Folder "{dirpath}" is cleared')


def cleardir(dirpath):
    for filename in os.listdir(dirpath):
        filepath = os.path.join(dirpath, filename)
        try:
            shutil.rmtree(filepath)
        except OSError:
            os.remove(filepath)


def do_dir(dirpath, clear=False):
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)
    elif clear:
        clear_files(dirpath)


def clear_files(dirpath):
    for filename in os.listdir(dirpath):
        filepath = os.path.join(dirpath, filename)
        if os.path.isfile(filepath):
            os.remove(filepath)
        elif os.path.isdir(filepath):
            clear_files(filepath)


def copy_dir(src, dst):
    try:
        shutil.copytree(src, dst)
    except Exception as e:
        logging.exception(e)
    sync_directory(dst)
    logging.info(f'The directory "{src}" is copied to "{dst}"')


def rem_dir(src):
    if not os.path.exists(src):
        return
    try:
        shutil.rmtree(src)
        logging.info(f'The directory "{src}" is deleted"')
    except Exception as e:
        logging.exception(e)


def move_dir(src, dst):
    try:
        shutil.move(src, dst)
        logging.info(f'The directory "{src}" is moved to "{dst}"')
    except Exception as e:
        logging.exception(e)

async def is_symlink_created(dest, timeout):
    step = 0.1
    total = 0.0
    while total < timeout:
        if os.path.islink(dest):
            return True
        total += step
        await asyncio.sleep(step)
    return False


async def create_symlink(orig, dest, timeout=1):
    try:
        os.symlink(os.path.abspath(orig), os.path.abspath(dest), target_is_directory=os.path.isdir(orig))
        res = await is_symlink_created(dest, timeout)
        logging.info(f'Symlink is created "{orig}" --> "{dest}"')
        return res
    except Exception as e:
        logging.exception(e)
    return False


async def link_package(src, dst, shared_folders):
    for entry in os.listdir(src):
        if entry in shared_folders:
            for entrance in os.listdir(os.path.join(src, entry)):
                symlink = f'{dst}/{entry}/{entrance}'
                if not is_symlink_exist(symlink):
                    await create_symlink(f'{src}/{entry}/{entrance}', symlink)
        else:
            await create_symlink(os.path.join(src, entry), os.path.join(dst, entry))


def is_symlink_exist(path):
    try:
        res = os.path.islink(os.path.abspath(path))
        return res
    except Exception as e:
        logging.exception(e)
    return False


def is_dir_exist(path):
    pth = Path(path)
    return pth.exists() and pth.is_dir()


def copy_file(orig, dest):
    try:
        shutil.copy(orig, dest)
    except Exception as e:
        logging.exception(e)
    with open(dest, 'rb') as f:
        f.flush()
        os.fsync(f.fileno())
    logging.info(f'The file "{orig}" is copied to "{dest}"')


def sync_directory(directory):
    for root, _, files in os.walk(directory):
        for file in files:
            file_path = os.path.join(root, file)
            with open(file_path, 'rb') as f:
                f.flush()
                os.fsync(f.fileno())


def compare_directories(dir1, dir2):
    def get_files_info(directory):
        files_info = {}
        for root, _, files in os.walk(directory):
            for file in files:
                file_path = os.path.join(root, file)
                if os.path.isfile(file_path):
                    file_size = os.path.getsize(file_path)
                    path = root[len(directory):] + '/' + file
                    files_info[path] = file_size
        return files_info

    return get_files_info(dir1) == get_files_info(dir2)


def adjust_path(path, process_name, shared_folders):
    pths = [pth for pth in sys.path]
    for pth in pths:
        if pth.startswith(path):
            sys.path.remove(pth)
    pth = f'{path}/work/{process_name}'
    makedir(pth, True)
    sys.path.append(pth)
    logging.info(f'SysPath "{pth}" is added')
    for folder in shared_folders:
        makedir(f'{pth}/{folder}')


def rename_dir(old_path, new_path):
    if not os.path.exists(old_path):
        return
    try:
        os.rename(old_path, new_path)
        logging.info(f'Folder "{old_path}" successfully renamed to "{new_path}"')
    except Exception as e:
        logging.exception(e)


def get_metadata(path):
    name = None
    ver = None
    dependencies = []
    try:
        with open(f'{path}/METADATA', 'r', encoding='utf-8') as f:
            for line in f:
                if line.startswith('Name:'):
                    name = line.split(':', 1)[1].strip() # .replace('-', '_')
                elif line.startswith('Version:'):
                    ver = line.split(':', 1)[1].strip()
                elif line.startswith('Requires-Dist:'):
                    dependency = version.parse_requires_dist(line.split('Requires-Dist:')[1].strip())
                    if is_right_dependency(dependency):
                        dependencies.append(dependency)
        return name, ver, dependencies
    except Exception as e:
        logging.exception(e)


def is_right_dependency(dependency):
    if 'extra' in dependency:
        return False
    if not version.check_version(sys.version, dependency.get('python_version')):
        return False
    if dependency.get('platform_system') and platform.system() != dependency.get('platform_system'):
        return False
    return True




import asyncio
import logging
import signal

from peacepie import params, msg_factory
from peacepie.assist import dir_opers
from peacepie.control.actors import actor_admin, actor_agent, actor_recreator, actor_seeker
from peacepie.control.intra import intra_link


ACTOR_ADMIN_COMMANDS = {'get_class', 'create_actor', 'create_actors', 'remove_actor',
                        'get_source_path', 'get_work_path',
                        'recreate_actor', 'create_replica', 'set_replica_params', 'transit_message', 'replica_resume'}

ACTOR_SEEKER_COMMANDS = {'seek_actor', 'find_actor'}

LIBS_PATH = 'libs_path'


class Admin:

    def __init__(self, lord, host_name, process_name, log_desc):
        self.is_head = False
        self.lord = lord
        self.host_name = host_name
        self.process_name = process_name
        self.log_desc = log_desc
        self.adaptor = None
        self.actor_admin = None
        self.asks = {}
        self.ask_index = 0
        self.cache = {}
        self.intralink = None
        self.actor_seeker = None

    def get_prefix(self):
        return f'{self.host_name}.{self.process_name}.'

    async def pre_run(self):
        self.actor_admin = actor_admin.ActorAdmin(self)
        self.actor_admin.actor_recreator = actor_recreator.ActorRecreator(self.actor_admin)
        asyncio.get_running_loop().create_task(self.actor_admin.actor_recreator.run())
        self.actor_admin.actor_agent = actor_agent.ActorAgent(self.actor_admin)
        asyncio.get_running_loop().create_task(self.actor_admin.actor_agent.run())
        if self.is_head:
            self.actor_seeker = actor_seeker.HeadActorSeeker(self)
        else:
            self.actor_seeker = actor_seeker.ActorSeeker(self)
        asyncio.get_running_loop().create_task(self.actor_seeker.run())
        self.intralink = intra_link.IntraLink(self)
        queue = asyncio.Queue()
        asyncio.get_running_loop().create_task(self.intralink.run(queue))
        await queue.get()

    async def quit(self):
        if self.adaptor.stop_event is not None:
            return
        self.adaptor.stop()

    async def exit(self):
        await self.actor_admin.exit()
        await self.intralink.exit()

    async def handle(self, msg):
        command = msg.get('command')
        body = msg.get('body') if msg.get('body') else {}
        sender = msg.get('sender')
        if command in ACTOR_ADMIN_COMMANDS:
            await self.actor_admin.handle(msg)
        elif command in ACTOR_SEEKER_COMMANDS:
            msg['recipient'] = self.actor_seeker.queue
            await self.adaptor.send(msg)
        elif command == 'remove_process':
            await self.remove_process()
        elif command == 'get_log_desc':
            ans = msg_factory.get_msg('log_desc', self.log_desc, sender)
            await self.adaptor.send(ans)
        elif command == 'change_cache':
            await self.change_cache(body, sender)
        elif command == 'remove_from_cache':
            await self.remove_from_cache(body, sender)
        elif command == 'get_members':
            await self.get_members(msg)
        else:
            return False
        return True

    async def remove_process(self):
        await self.quit()

    async def change_cache(self, body, recipient):
        await self.add_to_cache(body.get('node'), [body.get('entity')], True)
        if recipient:
            await self.adaptor.send(self.adaptor.get_msg('cache_is_changed', None, recipient))

    async def add_to_cache(self, node, names, is_exists = False):
        if node is None or self.adaptor.name == node:
            for name in names:
                if self.cache.get(name):
                    del self.cache[name]
            return
        queue = await self.intralink.get_intra_queue(node)
        if not queue:
            return
        for name in names:
            if is_exists and not self.cache.get(name):
                continue
            self.cache[name] = queue

    async def remove_from_cache(self, body, recipient):
        name = body.get('name')
        if name in self.cache:
            del self.cache[name]
        if recipient:
            await self.adaptor.send(self.adaptor.get_msg('removed_from_cache', None, recipient))

    async def get_members(self, msg):
        body = msg.get('body')
        page_size = body.get('page_size')
        level = body.get('level') if body.get('level') else 'prime'
        xid = body.get('id') if body.get('id') else ''
        page = int(xid.split('_')[2]) if xid.startswith('_page_') else 0
        members = []
        back = self.adaptor.name
        if level == 'actors':
            members = self.actor_admin.get_members()
            members = [{'next_level': 'actor', 'recipient': self.adaptor.name, 'id': member} for member in members]
            back = self.lord
        elif level == 'actor':
            members = [{'next_level': None, 'recipient': None, 'id': body.get('id')}]
        body = format_members(level, self.adaptor.name, page_size, page, members)
        body['_back'] = {'next_level': get_prev(level), 'recipient': back, 'id': '_back'}
        body['level'] = level
        ans = self.adaptor.get_msg('members', body, msg.get('sender'))
        await self.adaptor.send(ans)


def format_members(level, recipient, page_size, page, members):
    lng = len(members)
    count = lng // page_size + 1 if lng % page_size > 0 else lng // page_size
    nav = None
    if count > 1:
        members = members[page_size * page: page_size * (page + 1)]
        nav = {'next_level': level, 'recipient': recipient, 'count': count, 'page': page}
    res = {'members': members}
    if nav:
        res['nav'] = nav
    return res


def get_prev(level):
    if level == 'prime':
        return 'head'
    elif level == 'process':
        return 'prime'
    elif level == 'actors':
        return 'process'
    elif level == 'actor':
        return 'actors'
    return None



import asyncio
import re

from peacepie import adaptor


def get_alias(obj):
    performer = obj.performer if isinstance(obj, adaptor.Adaptor) else obj
    perf = performer
    res = None
    while True:
        if hasattr(perf, 'adaptor') and perf.adaptor and type(perf.adaptor) is adaptor.Adaptor:
            res = f'{perf.__class__.__name__} "{perf.adaptor.name}"' + (f' ({res})' if res else f'({id(perf)})')
            break
        if hasattr(perf, 'parent') and perf.parent is not None:
            if hasattr(perf, 'name'):
                res = f'"{perf.name}"' + (f'({res})' if res else '')
            else:
                res = perf.__class__.__name__ + (f'({res})' if res else '')
            perf = perf.parent
        else:
            name = perf.__class__.__name__
            if hasattr(perf, 'name'):
                name += f' "{perf.name}"'
            res = f'{name}' + (f' ({res})' if res else '')
            break
    return res


def async_sent_log(sender, msg):
    return get_alias(sender) + f' sent: {msg_format(msg)}'


def async_ask_log(sender, msg):
    return get_alias(sender) + f' asked: {msg_format(msg)}'


def async_received_log(sender, msg):
    return get_alias(sender) + f' received: {msg_format(msg)}'


def sync_sent_log(sender, msg):
    return get_alias(sender) + f' SENT: {msg_format(msg)}'


def sync_ask_log(sender, msg):
    return get_alias(sender) + f' ASKED: {msg_format(msg)}'


def sync_received_log(sender, msg):
    return get_alias(sender) + f' RECEIVED: {msg_format(msg)}'


def msg_format(msg):
    res = "{'mid': '" + msg.get('mid') + "', 'command': '" + str(msg.get('command'))
    res += "', 'body': " + body_format(msg.get('body')) + ", 'recipient': " + addr_format(msg.get('recipient'))
    res += ", 'sender': " + addr_format(msg.get('sender')) + ", 'timeout': " + str(msg.get('timeout'))
    res += "}"
    return res


pattern = r"'password':\s*(['\"])(.*?)(\1)"


def body_format(body):
    res = str(body)
    return re.sub(pattern, r"'password': '******'", res)


def addr_format(addr):
    if isinstance(addr, str):
        return f"'{addr}'"
    elif isinstance(addr, asyncio.Queue):
        return f'{addr.__class__.__name__}({id(addr)})'
    else:
        return str(addr)



import asyncio
import logging

from peacepie import loglistener, msg_factory, params
from peacepie.assist import dir_opers, log_util, misc
from peacepie.control import admin, delivery, package_loader, process_admin

PACKAGE_LOADER_COMMANDS = {'load_package'}

DELIVERY_COMMANDS = {'deliver_package', 'transfer'}


class PrimeAdmin(admin.Admin):

    def __init__(self, host_name, process_name):
        super().__init__(None, host_name, process_name, loglistener.instance.get_log_desc())
        self.process_admin = None
        self.package_loader = None
        self.delivery = None
        dir_opers.makedir(f'{params.instance["package_dir"]}/work', clear=True)

    async def pre_run(self):
        await super().pre_run()
        self.process_admin = process_admin.ProcessAdmin(self)
        self.package_loader = package_loader.PackageLoader(self)
        queue = asyncio.Queue()
        asyncio.get_running_loop().create_task(self.package_loader.run(queue))
        await queue.get()
        self.delivery = delivery.Delivery(self)
        asyncio.get_running_loop().create_task(self.delivery.run())

    async def quit(self):
        await self.process_admin.exit()
        self.adaptor.stop()

    async def exit(self):
        await self.actor_admin.exit()
        await self.intralink.exit()
        loglistener.instance.exit()

    async def handle(self, msg):
        command = msg.get('command')
        sender = msg.get('sender')
        if command == 'create_process':
            await self.create_process(sender)
        elif command == 'remove_process':
            await self.remove_process(msg)
        elif command == 'quiting':
            await self.quit()
        elif command in PACKAGE_LOADER_COMMANDS:
            await self.package_loader.queue.put(msg)
            logging.debug(log_util.async_sent_log(self, msg))
        elif command in DELIVERY_COMMANDS:
            msg['recipient'] = self.delivery.queue
            await self.adaptor.send(self, msg)
        elif command == 'get_members':
            await self.get_members(msg)
        else:
            return await super().handle(msg)
        return True

    async def create_process(self, recipient):
        asyncio.get_running_loop().create_task(self.process_admin.create_process(recipient))

    async def remove_process(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        complex_name = misc.ComplexName.parse_name(body.get('node'))
        recipient = msg.get('sender')
        asyncio.get_running_loop().create_task(self.process_admin.remove_process(complex_name))
        if recipient:
            await self.adaptor.send(msg_factory.get_msg('process_is_removed', None, recipient))

    async def get_members(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        page_size = body.get('page_size')
        level = body.get('level') if body.get('level') else 'prime'
        xid = body.get('id') if body.get('id') else ''
        page = int(xid.split('_')[2]) if xid.startswith('_page_') else 0
        members = []
        back = self.adaptor.name
        if level == 'process':
            members = self.process_admin.get_members()
            members = [{'next_level': 'actors', 'recipient': member, 'id': member} for member in members]
            back = self.adaptor.get_head_name()
        elif level == 'actors':
            members = self.actor_admin.get_members()
            members = [{'next_level': 'actor', 'recipient': self.adaptor.name, 'id': member} for member in members]
        elif level == 'actor':
            members = [{'next_level': None, 'recipient': None, 'id': body.get('id')}]
        body = admin.format_members(level, self.adaptor.name, page_size, page, members)
        body['_back'] = {'next_level': admin.get_prev(level), 'recipient': back, 'id': '_back'}
        body['level'] = level
        ans = self.adaptor.get_msg('members', body, msg.get('sender'))
        await self.adaptor.send(ans)



import asyncio
import importlib
import importlib.util
import inspect
import logging
import os

from peacepie.assist import (auxiliaries, log_util, json_util, serialization, dir_opers, terminal_util,
                             thread_util, timer)
from peacepie import msg_factory, params
from peacepie.control import ticker_admin, series_admin
from peacepie.control.head_prime_admin import HeadPrimeAdmin
from peacepie.control.admin import Admin


ADAPTOR_COMMANDS = {'quit', 'subscribe', 'unsubscribe', 'not_log_commands_set', 'not_log_commands_remove',
                    'cumulative_commands_set', 'cumulative_commands_remove', 'cumulative_tick',
                    'update_running', 'empty'}

CACHE_COMMANDS = {'actor_is_created', 'actors_are_created', 'actor_found'}


class Adaptor:

    def __init__(self, class_desc, name, parent, performer, sender=None):
        self.name = name
        self.parent = parent
        self.admin = parent if parent else performer
        self.sender = sender
        self.class_desc = class_desc
        self.performer = performer
        if not hasattr(self.performer, 'adaptor'):
            txt = f'The performer "{name}" does not have the attribute "adaptor"'
            raise AttributeError(txt)
        self.performer.adaptor = self
        self.msg = None
        self.pause_event = None
        self.idle_event = None
        self.is_idle = True
        self.stop_event = None
        self.queue = None
        self.ticker_admin = None
        self.observers = {}
        self.not_log_commands = set()
        self.cumulative_commands = {}
        self.cumulative_period = 10
        logging.info(f'{self.get_alias(self)} is created')

    def get_alias(self, obj=None):
        if not obj:
            obj = self
        return log_util.get_alias(obj)

    def pause(self):
        if self.pause_event is None:
            self.pause_event = asyncio.Event()
        if not self.is_idle:
            self.idle_event = asyncio.Event()
        if self.queue.empty():
            self.queue.put_nowait(msg_factory.get_msg('empty'))
        logging.info(f'{self.get_alias(self)} is paused')

    def is_paused(self):
        return self.pause_event is not None

    def resume(self, is_stopping=False):
        if is_stopping and self.stop_event is None:
            self.stop_event = asyncio.Event()
        if self.pause_event is not None:
            self.pause_event.set()
        logging.info(f'{self.get_alias(self)}(is_stopping={is_stopping}) is resumed')

    def stop(self):
        if self.pause_event is not None:
            self.pause_event.set()
        if self.stop_event is None:
            self.stop_event = asyncio.Event()
        if self.queue.empty():
            self.queue.put_nowait(msg_factory.get_msg('empty'))

    async def is_idling(self, timeout):
        if self.is_idle:
            return True
        res, _ = await asyncio.wait([asyncio.create_task(self.idle_event.wait())], timeout=timeout)
        self.idle_event = None
        return res

    async def is_stopped(self, timeout):
        res, _ = await asyncio.wait([asyncio.create_task(self.stop_event.wait())], timeout=timeout)
        return res

    async def run(self):
        if self.queue is None:
            self.queue = asyncio.Queue()
            if hasattr(self.performer, 'pre_run'):
                try:
                    await self.performer.pre_run()
                except Exception as ex:
                    logging.exception(ex)
            await self.is_running_notification()
        while True:
            self.is_idle = True
            try:
                if self.idle_event is not None:
                    self.idle_event.set()
                if self.pause_event is not None:
                    await self.pause_event.wait()
                    self.pause_event = None
                if self.stop_event is not None:
                    break
                self.is_idle = False
                self.msg = await self.queue.get()
                if not self.msg:
                    continue
                command = self.msg.get('command')
                if command not in self.not_log_commands:
                    if command in self.cumulative_commands.keys():
                        self.cumulative_commands[command]['received'] += 1
                    else:
                        logging.debug(log_util.async_received_log(self.performer, self.msg))
                if command in ADAPTOR_COMMANDS:
                    if not await self.handle(self.msg):
                        logging.warning(self.get_alias() + ' The message is not handled: ' + str(self.msg))
                        if self.msg.get('sender'):
                            ans = self.get_msg('is_not_handled', {'mid': self.msg.get('mid')},
                                               recipient=self.msg.get('sender'))
                            await self.send(ans)
                    continue
                if await self.performer.handle(self.msg):
                    if command in CACHE_COMMANDS:
                        await self.add_to_cache(self.msg)
                    await self.notify(self.msg)
                else:
                    logging.warning(self.get_alias() + ' The message is not handled: ' + str(self.msg))
                    if self.msg.get('sender'):
                        ans = self.get_msg('is_not_handled', {'mid': self.msg.get('mid')},
                                           recipient=self.msg.get('sender'))
                        await self.send(ans)
            except asyncio.exceptions.CancelledError:
                pass
            except BaseException as ex:
                logging.exception(ex)
        await self.exit()
        if self.stop_event is not None:
            self.stop_event.set()
        logging.info(f'{self.get_alias(self)} is stopped')

    async def exit(self):
        if not hasattr(self.performer, 'exit'):
            return
        if isinstance(self.performer, Admin):
            await self.performer.exit()
            return
        timeout = 4
        if hasattr(self.performer, 'shield_timeout') and self.performer.shield_timeout > 0:
            timeout =self.performer.shield_timeout
            task = asyncio.shield(self.performer.exit())
        else:
            task = asyncio.create_task(self.performer.exit())
        try:
            await asyncio.wait_for(task, timeout=timeout)
        except asyncio.TimeoutError:
            task.cancel()
        except Exception as e:
            logging.exception(e)

    async def is_running_notification(self):
        if not self.sender:
            return
        system_name = self.get_param('system_name')
        node = self.parent.adaptor.name if self.parent else self.name
        entity = self.name if self.parent else None
        body = self.get_addr(system_name, node, entity)
        msg = msg_factory.get_msg('actor_is_created', body, recipient=self.sender)
        await self.send(msg)

    async def handle(self, msg):
        command = msg.get('command')
        body = msg.get('body') if msg.get('body') else {}
        sender = msg.get('sender')
        if command == 'quit':
            await self.quit(msg)
        elif command == 'cumulative_tick':
            self.cumulative_tick()
        elif command == 'cumulative_commands_set':
            await self.cumulative_commands_set(body.get('commands'), sender)
        elif command == 'cumulative_commands_remove':
            await self.cumulative_commands_remove(body.get('commands'), sender)
        elif command == 'not_log_commands_set':
            await self.not_log_commands_set(body.get('commands'), sender)
        elif command == 'not_log_commands_remove':
            await self.not_log_commands_remove(body.get('commands'), sender)
        elif command == 'subscribe':
            self.subscribe(body.get('command'), sender)
        elif command == 'unsubscribe':
            self.unsubscribe(body.get('command'), sender)
        elif command == 'empty':
            pass
        else:
            return False
        return True

    async def quit(self, msg):
        if isinstance(self.performer, HeadPrimeAdmin):
            if msg.get('sender'):
                await self.send(self.get_msg('is_quiting', None, msg.get('sender')))
            await asyncio.sleep(0.4)
            asyncio.create_task(self.performer.quit())
        else:
            msg['recipient'] = self.get_head_addr()
            await self.send(msg)

    def cumulative_tick(self):
        for command in self.cumulative_commands.keys():
            count = self.cumulative_commands[command]['received']
            logging.info(f'{self.get_alias(self)} received {count} "{command}" messages')
            count = self.cumulative_commands[command]['local_sent']
            if count > 0:
                logging.info(f'{self.get_alias(self)} sent {count} "{command}" messages')
            count = self.cumulative_commands[command]['remote_sent']
            if count > 0:
                logging.info(f'{self.get_alias(self)} SENT {count} "{command}" MESSAGES')
            count = self.cumulative_commands[command]['local_asked']
            if count > 0:
                logging.info(f'{self.get_alias(self)} asked {count} "{command}" messages')
            count = self.cumulative_commands[command]['remote_asked']
            if count > 0:
                logging.info(f'{self.get_alias(self)} ASKED {count} "{command}" MESSAGES')
            self.cumulative_commands[command]['received'] = 0
            self.cumulative_commands[command]['local_sent'] = 0
            self.cumulative_commands[command]['remote_sent'] = 0
            self.cumulative_commands[command]['local_asked'] = 0
            self.cumulative_commands[command]['remote_asked'] = 0

    async def cumulative_commands_set(self, commands, recipient=None):
        for command in commands:
            val = {'received': 0, 'local_sent': 0, 'remote_sent': 0, 'local_asked': 0, 'remote_asked': 0}
            self.cumulative_commands[command] = val
        name = 'cumulative'
        if not self.ticker_admin or not self.ticker_admin.is_ticker_exists(name):
            self.add_ticker(self.cumulative_period, command='cumulative_tick', name=name)
        if recipient:
            await self.send(self.get_msg('set', recipient=recipient))

    async def cumulative_commands_remove(self, commands, recipient=None):
        for command in commands:
            del self.cumulative_commands[command]
        if not self.cumulative_commands:
            self.ticker_admin.remove('cumulative')
        if recipient:
            await self.send(self.get_msg('removed', recipient=recipient))

    async def not_log_commands_set(self, commands, recipient=None):
        self.not_log_commands.update(commands)
        if recipient:
            await self.send(self.get_msg('set', recipient=recipient))

    async def not_log_commands_remove(self, commands, recipient=None):
        for command in commands:
            del self.cumulative_commands[command]
        if recipient:
            await self.send(self.get_msg('removed', recipient=recipient))

    def subscribe(self, command, sender):
        res = self.observers.get(command)
        if not res:
            res = []
            self.observers[command] = res
        res.append(sender)

    def unsubscribe(self, command, sender):
        res = self.observers.get(command)
        if res:
            res.remove(sender)

    async def notify(self, msg):
        res = self.observers.get(msg.get('command'))
        if not res:
            return
        for recipient in res:
            await self.send(msg_factory.get_msg('notification', msg, recipient=recipient))

    def json_loads(self, jsn):
        return json_util.json_loads(jsn)

    def json_dumps(self, jsn):
        return json_util.json_dumps(jsn)

    def get_msg(self, command, body=None, recipient=None, sender=None, timeout=None):
        return msg_factory.get_msg(command, body, recipient, sender, timeout)

    def get_control_msg(self, command, body=None, recipient=None, sender=None):
        return msg_factory.get_control_msg(command, body, recipient, sender)

    def add_ticker(self, period, delay=None, count=None, name=None, command=None):
        if not self.ticker_admin:
            self.ticker_admin = ticker_admin.TickerAdmin()
        if self.ticker_admin.is_ticker_exists(name):
            return name
        return self.ticker_admin.add_ticker(self.queue, period, delay, count, name, command)

    def remove_ticker(self, name):
        if not self.ticker_admin:
            return
        self.ticker_admin.remove_ticker(name)

    def get_param(self, param_name):
        return params.instance.get(param_name)

    def get_test_param(self, param_name):
        return params.test_instance.get(param_name)

    def get_serializer(self):
        return serialization.Serializer()

    def get_serializer_desc(self):
        return f"'{serialization.__name__}|{serialization.Serializer.__name__}'"

    def get_serializer_spec(self):
        path, module = os.path.split(inspect.getmodule(serialization).__file__)
        return {'path': path, 'module': os.path.splitext(module)[0], 'class': serialization.Serializer.__name__}

    def get_package_path(self, path_only:bool=True):
        res = None
        spec = importlib.util.find_spec(__package__)
        if spec and spec.origin:
            res = os.path.dirname(spec.origin)
            if path_only:
                res = os.path.dirname(res)
            res = f"'{res}'"
        return res

    def makedir(self, dirpath, clear=False):
        dir_opers.makedir(dirpath, clear)

    def execute(self, cmd):
        return terminal_util.execute(cmd)

    def sync_as_async(self, sync_function, sync_args=None):
        return thread_util.sync_as_async(sync_function, sync_args)

    async def com_exe(self, coms, timeout=300):
        if isinstance(coms, str):
            coms = ([coms], timeout)
        elif isinstance(coms, list):
            coms = (coms, timeout)
        res = await self.sync_as_async(self.execute, sync_args=coms)
        if res[0] == 0:
            logging.debug(f'{coms}: ({res[0]}, {self.res_squeeze(res[1])}, {self.res_squeeze(res[2])})')
        else:
            raise Exception(f'{coms}: ({res[0]}, {self.res_squeeze(res[1])}, {self.res_squeeze(res[2])})')
        return res

    async def command_execute(self, command, command_for_log, timeout=300):
        res = None
        try:
            process = await asyncio.create_subprocess_shell(
                command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
            )
            await asyncio.wait_for(process.wait(), timeout)
            stdout, stderr = await process.communicate()
            res = (process.returncode, stdout.decode('utf-8'), stderr.decode('utf-8'))
            title = f'{self.get_alias()} executed command "{command_for_log}"'
            text = f'{title}: ({res[0]}, "{self.res_squeeze(res[1])}", "{self.res_squeeze(res[2])}")'
            if res[0] == 0:
                logging.debug(text)
            else:
                logging.warning(text)
        except Exception as e:
            logging.exception(e)
        return res

    async def sudo_command_execute(self, command, credentials_name, timeout=300):
        credentials = await self.get_credentials(credentials_name)
        username = credentials.get('username')
        password = credentials.get('password')
        com = f'echo "{password}" | su -c "echo {password} | sudo -S {command}" - {username}'
        password = '******'
        com_for_log = f'echo "{password}" | su -c "echo {password} | sudo -S {command}" - {username}'
        res = await self.command_execute(com, com_for_log, timeout)
        return res[0] == 0

    async def sudo_com_exe(self, credentials_name, command, cwd=None, timeout=300):
        credentials = await self.get_credentials(credentials_name)
        username = credentials.get('username')
        password = credentials.get('password')
        com = f'echo "{password}" | su -c "echo {password} | sudo -S {command}" - {username}'
        password = '******'
        com_for_log = f'echo "{password}" | su -c "echo {password} | sudo -S {command}" - {username}'
        res = await self.sync_as_async(terminal_util.exe, sync_args=(com, cwd, timeout))
        title = f'{self.get_alias()} executed command "{com_for_log}"'
        text = f'{title}: ({res[0]}, "{self.res_squeeze(res[1])}", "{self.res_squeeze(res[2])}")'
        if res[0] == 0:
            logging.debug(text)
        else:
            logging.warning(text)
        return res[0] == 0

    def res_squeeze(self, param):
        res = param[:200] + ' >>>>' if len(param) > 200 else param
        return res.replace("\n", "\\n")

    def series_next(self, name):
        return series_admin.instance.next(name)

    def series_set_value(self, name, value):
        series_admin.instance.set_value(name, value)

    def series_set_max_value(self, name, max_value):
        series_admin.instance.set_max_value(name, max_value)

    def get_caller_info(self):
        caller_frame = inspect.currentframe().f_back
        caller_info = inspect.getframeinfo(caller_frame)
        file_name = caller_info.filename
        module_name = inspect.getmodulename(file_name)
        line_number = caller_info.lineno
        package_name = None
        module = inspect.getmodule(caller_frame)
        if module and hasattr(module, '__package__'):
            package_name = module.__package__
        return f'Package: {package_name}, Module: {module_name}, Line: {line_number}'

    def start_timer(self, timeout, queue=None, mid=None):
        if not queue:
            queue = self.queue
        timer.start(timeout, queue, mid)

    def get_head_name(self):
        if self.parent:
            if self.parent.intralink.head:
                return self.parent.intralink.head
            else:
                return self.parent.adaptor.name
        else:
            if self.performer.intralink.head:
                return self.performer.intralink.head
            else:
                return self.name

    def get_prime_name(self):
        if self.parent:
            if self.parent.lord:
                return self.parent.lord
            else:
                return self.parent.adaptor.name
        else:
            return self.name

    def get_addr(self, system, node, entity):
        if system:
            return {'system': system, 'node': node, 'entity': entity}
        else:
            return {'node': node, 'entity': entity}

    def add_system(self, addr, system_name):
        return self.get_addr(system_name, addr['node'], addr['entity'])

    def get_head_addr(self):
        return self.get_addr(None, self.get_head_name(), None)

    def get_prime_addr(self):
        return self.get_addr(None, self.get_prime_name(), None)

    def get_self_addr(self):
        if self.parent:
            return self.get_addr(None, self.parent.adaptor.name, self.name)
        else:
            return self.get_addr(None, self.name, None)

    def get_node(self):
        if self.parent:
            return self.parent.adaptor.name
        else:
            return self.name

    async def get_all_nodes(self):
        if hasattr(self.performer, 'get_all_nodes'):
            return self.performer.get_all_nodes()
        ans = await self.ask(self.get_msg('get_all_nodes', None, self.get_head_name()))
        if ans.get('command') == 'all_nodes':
            return ans.get('body')
        return None

    async def clarify_recipient(self, recipient):
        if recipient is None:
            return self.admin.adaptor.queue
        if type(recipient) is dict:
            system_name = recipient.get('system')
            if system_name and system_name != params.instance.get('system_name'):
                head = self.get_head_name()
                if head == self.admin.adaptor.name:
                    return self.admin.interlink.queue
                else:
                    return await self.admin.intralink.get_intra_queue(head)
            if recipient.get('node') == self.admin.adaptor.name:
                if recipient.get('entity'):
                    res = self.admin.actor_admin.get_actor_queue(recipient.get('entity'))
                    if res:
                        return res
                    recipient = recipient.get('entity')
                else:
                    return self.admin.adaptor.queue
            else:
                return await self.admin.intralink.get_intra_queue(recipient['node'])
        if type(recipient) is str:
            if recipient == self.admin.adaptor.name:
                return self.admin.adaptor.queue
            elif recipient.startswith('_'):
                return self.admin.asks.get(recipient)
            else:
                res = self.admin.cache.get(recipient)
                if not res:
                    res = self.admin.actor_admin.get_actor_queue(recipient)
                if not res:
                    res = await self.admin.intralink.get_intra_queue(recipient)
                return res
        else:
            return recipient

    async def send(self, msg, sender=None):
        if not sender:
            sender = self
        recipient = msg.get('recipient')
        res = await self.clarify_recipient(recipient)
        if not res:
            res = await self.find(sender, recipient)
            if not res:
                return
        is_local = isinstance(res, asyncio.Queue)
        await res.put(msg)
        self.send_log(sender, is_local, msg)

    def send_log(self, sender, is_local, msg):
        command = msg.get('command')
        if command in sender.not_log_commands:
            return
        if command in sender.cumulative_commands.keys():
            sender.cumulative_commands[command]['local_sent' if is_local else 'remote_sent'] += 1
        elif is_local:
            logging.debug(log_util.async_sent_log(sender, msg))
        else:
            logging.debug(log_util.sync_sent_log(sender, msg))

    async def ask(self, msg, timeout=1, questioner=None):
        if not questioner:
            questioner = self
        recipient = msg.get('recipient')
        res = await self.clarify_recipient(recipient)
        if not res:
            res = await self.find(questioner, recipient)
            if not res:
                return None
        to = msg.get('timeout')
        if to and to > timeout:
            timeout = to
        else:
            msg['timeout'] = timeout
        queue = asyncio.Queue()
        entity = f'_{self.admin.ask_index}'
        self.admin.ask_index += 1
        msg['sender'] = {'node': self.admin.adaptor.name, 'entity': entity}
        self.admin.asks[entity] = queue
        if not (isinstance(recipient, str) or isinstance(recipient, dict)):
            msg['recipient'] = None
        await res.put(msg)
        self.ask_log(questioner, isinstance(res, asyncio.Queue), msg)
        timer.start(timeout, queue, msg['mid'])
        ans = await queue.get()
        if ans.get('command') == 'timer':
            logging.warning(log_util.async_received_log(questioner, ans))
        else:
            self.answer_on_ask_log(questioner, ans)
        del self.admin.asks[entity]
        if ans.get('command') in CACHE_COMMANDS:
            await self.add_to_cache(ans)
        return ans

    async def group_ask(self, timeout, count, get_values, questioner=None):
        if not questioner:
            questioner = self
        entity = f'_{self.admin.ask_index}'
        self.admin.ask_index += 1
        queue = asyncio.Queue()
        self.admin.asks[entity] = queue
        sender = {'node': self.admin.adaptor.name, 'entity': entity}
        waiting_count = count
        group_mid = msg_factory.get_group_mid()
        timer.start(timeout, queue, group_mid)
        for index in range(count):
            values = get_values(index)
            recipient = values.get('recipient')
            res = await self.clarify_recipient(recipient)
            if not res:
                res = await self.find(questioner, recipient)
                if not res:
                    waiting_count -= 1
                    continue
            recipient = recipient if isinstance(recipient, str) or isinstance(recipient, dict) else None
            command = values.get('command')
            body = values.get('body')
            msg = msg_factory.get_msg(command, body, recipient=recipient, sender=sender, timeout=timeout,
                                      group_mid=group_mid)
            await res.put(msg)
            self.ask_log(questioner, isinstance(res, asyncio.Queue), msg)
        while waiting_count > 0:
            ans = await queue.get()
            waiting_count -= 1
            if ans.get('command') == 'timer':
                del self.admin.asks[entity]
                return ans
            else:
                if ans.get('command') in CACHE_COMMANDS:
                    await self.add_to_cache(ans)
                self.answer_on_ask_log(questioner, ans)
        del self.admin.asks[entity]
        return msg_factory.get_msg('group_ask_completed')

    def ask_log(self, questioner, is_local, msg):
        command = msg.get('command')
        if command in questioner.not_log_commands:
            return
        if command in questioner.cumulative_commands.keys():
            questioner.cumulative_commands[command]['local_asked' if is_local else 'remote_asked'] += 1
        elif is_local:
            logging.debug(log_util.async_ask_log(questioner, msg))
        else:
            logging.debug(log_util.sync_ask_log(questioner, msg))

    def answer_on_ask_log(self, questioner, msg):
        command = msg.get('command')
        if command in questioner.not_log_commands:
            return
        if command in questioner.cumulative_commands.keys():
            questioner.cumulative_commands[command]['received'] += 1
        else:
            logging.debug(log_util.async_received_log(questioner, msg))

    async def find(self, sender, entity):
        res = None
        msg = msg_factory.get_msg('seek_actor', {'entity': entity}, self.admin.adaptor.get_head_addr())
        ans = await self.ask(msg, 2, sender)
        if ans['command'] == 'actor_found':
            res = await self.admin.intralink.get_intra_queue(ans['body']['node'])
            if res:
                self.admin.cache[entity] = res
            else:
                logging.warning(f'The actor "{entity}" is not found')
        return res

    async def add_to_cache(self, msg):
        command = msg.get('command')
        body = msg.get('body')
        if not body:
            return
        system =body.get('system')
        if system and system != self.admin.adaptor.get_param('system_name'):
            return
        node = body.get('node')
        if not node or node == self.admin.adaptor.name:
            return
        if command == 'actors_are_created':
            names = body.get('names')
            if not names:
                return
        else:
            name = body.get('entity')
            if not name:
                return
            names = [name]
        await self.admin.add_to_cache(node, names)

    async def get_credentials(self, credentials_name):
        body = {'credentials_name': credentials_name}
        query = self.get_msg('get_credentials', body, self.get_head_addr())
        ans = await self.ask(query)
        body = ans.get('body') if isinstance(ans.get('body'), dict) else {}
        return body

    def get_log_path(self):
        res = None
        for handler in logging.getLogger().handlers:
            if isinstance(handler, logging.FileHandler):
                res = os.path.dirname(handler.baseFilename)
                break
        return res

    def get_current_time_string(self):
        return auxiliaries.get_current_time_string()

    def get_attributes(self):
        return [attr for attr in vars(self.performer)
                if not callable(getattr(self.performer, attr)) and not attr.startswith('_') and attr != 'adaptor']

    def get_params(self):
        return [{'name': attr, 'value': value} for attr, value in vars(self.performer).items()
                if not callable(getattr(self.performer, attr)) and not attr.startswith('_') and attr != 'adaptor']

    def set_params(self, params):
        names = self.get_attributes()
        for param in params:
            name = param.get('name')
            if name in names:
                setattr(self.performer, name, param.get('value'))

    def has_attr(self, attr_name, is_callable=False, is_coroutine=False):
        if not hasattr(self.performer, attr_name):
            return False
        if is_callable:
            method = getattr(self.performer, attr_name)
            if not callable(method):
                return False
            if is_coroutine:
                if not inspect.iscoroutinefunction(method):
                    return False
        return True



import asyncio
import logging

from peacepie import msg_factory
from peacepie.assist import log_util
from peacepie.control.actors import actor_loader, package_admin


LOADER_COMMANDS = {'get_class', 'create_actor', 'create_replica', 'create_actors'}
RECREATE_COMMANDS = {'recreate_actor'}
AGENT_COMMANDS = {'set_replica_params', 'transit_message', 'replica_resume'}


class ActorAdmin:

    def __init__(self, parent):
        self.parent = parent
        self.package_admin = package_admin.PackageAdmin(self)
        self.actor_recreator = None
        self.actor_agent = None
        self.actor_loaders = []
        self.actors = {}
        self.not_log_commands = set()
        self.cumulative_commands = {}
        logging.info(log_util.get_alias(self) + ' is created')

    async def exit(self):
        tasks = []
        stop_events = []
        for actor in self.actors.values():
            adaptor = actor.get('adaptor')
            adaptor.stop()
            stop_events.append(adaptor.stop_event.wait())
            tasks.append(actor.get('task'))
        try:
            await asyncio.wait_for(asyncio.gather(*stop_events), timeout=0.4)
        except asyncio.TimeoutError:
            pass
        await self.exiting(tasks)
        tasks = [t for t in asyncio.all_tasks()
                 if t is not asyncio.current_task() and t not in self.parent.intralink.tasks]
        if tasks:
            await self.exiting(tasks)

    async def exiting(self, tasks):
        [task.cancel() for task in tasks]
        try:
            await asyncio.wait_for(asyncio.gather(*tasks), timeout=0.4)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            logging.exception(e)

    async def handle(self, msg):
        command = msg.get('command')
        if command in LOADER_COMMANDS:
            for loader in self.actor_loaders:
                if loader['loader'].queue.qsize() < 10:
                    await loader['loader'].queue.put(msg)
                    logging.debug(log_util.async_sent_log(self, msg))
                    return True
            loader = self.add_actor_loader()
            await loader.queue.put(msg)
            logging.debug(log_util.async_sent_log(self, msg))
        elif command in RECREATE_COMMANDS:
            await self.actor_recreator.queue.put(msg)
            logging.debug(log_util.async_sent_log(self, msg))
        elif command in AGENT_COMMANDS:
            await self.actor_agent.queue.put(msg)
            logging.debug(log_util.async_sent_log(self, msg))
        elif command == 'remove_actor':
            await self.remove_actor(msg)
        elif command == 'get_source_path':
            body = {'path': self.package_admin.source_path}
            ans = self.parent.adaptor.get_msg('source_path', body, recipient=msg.get('sender'))
            await self.parent.adaptor.send(ans)
        elif command == 'get_work_path':
            body = {'path': self.package_admin.work_path}
            ans = self.parent.adaptor.get_msg('work_path', body, recipient=msg.get('sender'))
            await self.parent.adaptor.send(ans)
        else:
            return False
        return True

    def add_actor_loader(self):
        loader = actor_loader.ActorLoader(self)
        task = asyncio.get_running_loop().create_task(loader.run())
        self.actor_loaders.append({'loader': loader, 'task': task})
        return loader

    def get_actor_queue(self, name):
        actor = self.actors.get(name)
        if not actor:
            return None
        adaptor = actor.get('adaptor')
        if not adaptor:
            return None
        return adaptor.queue

    def get_members(self):
        res = [actor for actor in self.actors]
        res.sort()
        res.insert(0, self.parent.adaptor.name)
        return res

    async def remove_actor(self, msg):
        recipient = msg.get('sender')
        body = msg.get('body') if msg.get('body') else {}
        name = body.get('name')
        if not name:
            if recipient:
                await self.parent.adaptor.send(self.parent.adaptor.get_msg('actor_is_absent', body, recipient))
            return
        if await self.removing_actor(name) and recipient:
            await self.parent.adaptor.send(self.parent.adaptor.get_msg('actor_is_removed', body, recipient))
        elif recipient:
            await self.parent.adaptor.send(self.parent.adaptor.get_msg('actor_is_absent', {'name': name}, recipient))

    async def removing_actor(self, name):
        actor = None
        try:
            actor = self.actors.get(name)
        except Exception as e:
            logging.exception(e)
        if not actor:
            return False
        adaptor = actor.get('adaptor')
        alias = log_util.get_alias(adaptor)
        adaptor.stop()
        if not await adaptor.is_stopped(5):
            return False
        new_addr = {'node': None, 'entity': name}
        msg = msg_factory.get_msg('change_caches', new_addr, self.parent.adaptor.get_head_addr())
        if self.parent.is_head:
            await self.parent.change_caches(msg)
        else:
            await self.parent.adaptor.ask(msg, 10, self)
        del self.actors[name]
        logging.info(f'{alias} is removed')
        return True



import asyncio
import logging

from peacepie import msg_factory
from peacepie.assist import log_util, timer


class ActorRecreator:

    def __init__(self, parent):
        self.parent = parent
        self.grandparent = parent.parent
        self.queue = asyncio.Queue()
        self.queue_storage = {}
        self.not_log_commands = set()
        self.cumulative_commands = {}
        logging.info(log_util.get_alias(self) + ' is created')

    async def run(self):
        while True:
            msg = await self.queue.get()
            logging.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    logging.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                logging.exception(ex)

    async def handle(self, msg):
        command = msg.get('command')
        if command == 'recreate_actor':
            await self.recreate_actor(msg)
        else:
            return False
        return True

    async def recreate_actor(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        node = body.get('node')
        old_entity = body.get('entity')
        recipient = msg.get('sender')
        timeout = (msg.get('timeout') if msg.get('timeout') else 20) / 2
        is_locally = node == self.grandparent.adaptor.get_node()
        actor = self.parent.actors.get(old_entity)
        if not actor:
            await self.redirect(old_entity, recipient, msg)
            return
        adaptor = actor.get('adaptor')
        body = {'class_desc': adaptor.class_desc, 'name': None if is_locally else old_entity}
        ans = await self.grandparent.adaptor.ask(msg_factory.get_msg('create_replica', body, node), timeout, self)
        if ans.get('command') != 'actor_is_created':
            await self.failure(recipient)
            return
        adaptor.pause()
        new_entity = ans.get('body').get('entity')
        is_idling = await adaptor.is_idling(timeout)
        params = adaptor.get_params()
        try:
            if is_locally:
                self.transit_locally(is_idling, old_entity, new_entity, params)
            else:
                await self.transit_remotely(node, is_idling, new_entity, params, timeout)
        except Exception as e:
            logging.exception(e)
            self.parent.actors[old_entity] = actor
            adaptor.resume()
            return
        await self.stop(actor, timeout)
        if is_locally:
            self.parent.actors.get(old_entity).get('adaptor').resume()
        else:
            msg = msg_factory.get_msg('replica_resume', {'entity': old_entity}, {'node': node, 'entity': None})
            await self.grandparent.adaptor.ask(msg, timeout, self)
        if recipient:
            await self.grandparent.adaptor.send(msg_factory.get_msg('actor_is_recreated', None, recipient), self)

    async def redirect(self, entity, recipient, msg):
        queue = self.grandparent.cache.get(entity)
        if queue:
            msg['recipient'] = None
            await queue.put(msg)
            logging.debug(log_util.sync_sent_log(self, msg))
            return
        ans = await self.grandparent.adaptor.ask(msg_factory.get_msg('seek_actor', {'entity': entity}))
        if ans.get('command') != 'actor_found':
            await self.failure(recipient)
            return
        msg['recipient'] = {'node': ans.get('body').get('node'), 'entity': None}
        await self.grandparent.adaptor.send(msg, self)

    async def failure(self, recipient):
        if not recipient:
            return
        await self.grandparent.adaptor.send(msg_factory.get_msg('actor_is_not_recreated', None, recipient), self)

    def transit_locally(self, is_idling, old_entity, new_entity, params):
        old_actor = self.parent.actors.pop(old_entity)
        old_adaptor = old_actor.get('adaptor')
        new_actor = self.parent.actors.pop(new_entity)
        new_adaptor = new_actor.get('adaptor')
        new_adaptor.name = old_entity
        performer = new_adaptor.performer
        logging.info(f'{performer.__class__.__name__} "{new_entity}"({id(performer)}) has renamed to "{old_entity}"')
        new_adaptor.set_params(params)
        try:
            new_adaptor.queue.put_nowait(msg_factory.get_msg('start'))
            if not is_idling:
                new_adaptor.queue.put_nowait(old_adaptor.msg)
            while True:
                msg = old_adaptor.queue.get_nowait()
                new_adaptor.queue.put_nowait(msg)
        except asyncio.QueueEmpty:
            pass
        if old_adaptor.not_log_commands:
            new_adaptor.not_log_commands.update(old_adaptor.not_log_commands)
        self.parent.actors[old_entity] = new_actor

    async def transit_remotely(self, node, is_idling, entity, params, timeout):
        body = {'entity': entity, 'params': params}
        msg = msg_factory.get_msg('set_replica_params', body, {'node': node, 'entity': None})
        await self.grandparent.adaptor.ask(msg, timeout, self)
        await self.grandparent.add_to_cache(node, [entity])
        new_addr = {'node': node, 'entity': entity}
        msg = msg_factory.get_msg('change_caches', new_addr, self.grandparent.adaptor.get_head_addr())
        await self.grandparent.adaptor.ask(msg, 10, self)
        actor = self.parent.actors.pop(entity)
        adaptor = actor.get('adaptor')
        if adaptor.not_log_commands:
            msg = msg_factory.get_msg('not_log_commands_set', {'commands': list(adaptor.not_log_commands)})
            await self.send_to_clone(node, entity, msg)
        await self.send_to_clone(node, entity, msg_factory.get_msg('start'))
        if not is_idling:
            await self.send_to_clone(node, entity, adaptor.msg)
        while not adaptor.queue.empty():
            await self.send_to_clone(node, entity, await adaptor.queue.get())

    async def send_to_clone(self, node, entity, msg):
        res = msg_factory.get_msg('transit_message', {'entity': entity, 'message': msg}, {'node': node, 'entity': None})
        await self.grandparent.adaptor.send(res, self)

    async def stop(self, actor, timeout):
        adaptor = actor.get('adaptor')
        adaptor.resume(True)
        if not await adaptor.is_stopped(timeout):
            try:
                task = actor.get('task')
                task.cancel()
                await asyncio.wait_for(task, timeout)
            except Exception as e:
                logging.exception(e)



import asyncio
import logging

from peacepie import msg_factory
from peacepie.assist import log_util, timer


class ActorSeeker:

    def __init__(self, parent):
        self.logger = logging.getLogger()
        self.parent = parent
        self.queue = asyncio.Queue()
        self.not_log_commands = set()
        self.cumulative_commands = {}
        self.logger.info(log_util.get_alias(self) + ' is created')

    async def run(self):
        while True:
            msg = await self.queue.get()
            self.logger.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    self.logger.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                self.logger.exception(ex)

    async def handle(self, msg):
        command = msg['command']
        if command == 'seek_actor':
            await self.seek_actor(msg)
        elif command == 'find_actor':
            await self.find_actor(msg)
        else:
            return False
        return True

    async def seek_actor(self, msg):
        if await self.find_actor(msg):
            return
        msg['recipient'] = {'node': self.parent.intralink.head, 'entity': None}
        await self.parent.adaptor.send(msg, self)

    async def find_actor(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        entity = body.get('entity')
        res = self.parent.actor_admin.actors.get(entity)
        if not res:
            return False
        body = {'node': self.parent.adaptor.name, 'entity': entity}
        message = msg_factory.get_msg('actor_found', body, recipient=msg['sender'])
        await self.parent.adaptor.send(message, self)
        return True


class HeadActorSeeker:

    def __init__(self, parent):
        self.logger = logging.getLogger()
        self.parent = parent
        self.queue = asyncio.Queue()
        self.not_log_commands = set()
        self.cumulative_commands = {}
        self.logger.info(log_util.get_alias(self) + ' is created')

    async def run(self):
        while True:
            msg = await self.queue.get()
            self.logger.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    self.logger.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                self.logger.exception(ex)

    async def handle(self, msg):
        command = msg['command']
        if command == 'seek_actor':
            await self.seek_actor(msg)
        elif command == 'find_actor':
            await self.find_actor(msg)
        else:
            return False
        return True

    async def seek_actor(self, msg):
        if await self.find_actor(msg):
            return
        queue = asyncio.Queue()
        entity = f'_{self.parent.ask_index}'
        self.parent.ask_index += 1
        self.parent.asks[entity] = queue
        sender = {'node': self.parent.adaptor.name, 'entity': entity}
        message = msg_factory.get_msg('find_actor', msg['body'], sender=sender)
        for recipient in self.parent.intralink.get_recipients():
            await recipient.put(message)
            self.logger.debug(log_util.sync_ask_log(self, message))
        timer.start(1, queue, message['mid'])
        res = await queue.get()
        del self.parent.asks[entity]
        if res['command'] == 'tick':
            return
        res['recipient'] = msg['sender']
        await self.parent.adaptor.send(res, self)

    async def find_actor(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        entity = body.get('entity')
        res = self.parent.actor_admin.actors.get(entity)
        if not res:
            return False
        body = {'node': self.parent.adaptor.name, 'entity': entity}
        message = msg_factory.get_msg('actor_found', body, recipient=msg['sender'])
        await self.parent.adaptor.send(message, self)
        return True



import logging

from peacepie import params, msg_factory
from peacepie.assist import log_util, json_util


def load_credentials():
    res = None
    try:
        with open(params.instance.get('safe_config'), 'r') as file:
            json = file.read()
        res = json_util.json_loads(json)
    except Exception as e:
        logging.exception('safe_config_read_error', e)
    return res


class SafeAdmin:

    def __init__(self, parent):
        self.parent = parent
        self.credentials = load_credentials()
        self.not_log_commands = set()
        self.cumulative_commands = {}
        logging.info(log_util.get_alias(self) + ' is created')

    async def handle(self, msg):
        logging.debug(log_util.async_received_log(self, msg))
        command = msg.get('command')
        if command == 'get_credentials':
            body = msg.get('body') if isinstance(msg.get('body'), dict) else {}
            body = self.credentials.get(body.get('credentials_name'))
            ans = msg_factory.get_msg('credentials', body, msg.get('sender'))
            await self.parent.adaptor.send(ans, self)
        else:
            logging.warning(log_util.get_alias(self) + ' The message is not handled: ' + str(msg))



import multiprocessing
import os

import json
import logging
import logging.config
import sys
from logging.handlers import RotatingFileHandler

from peacepie import params
from peacepie.assist import dir_opers
from peacepie.assist.auxiliaries import is_pycharm

LOG_PATH = 'logs/log.log'

logger = None
# logger_listener = None
# log_desc = None


def logger_start(config_filename):
    global logger
    # global logger_listener
    # global log_desc
    if logger:
        return
    process = ''
    if params.instance.get('separate_log_per_process'):
        process = f'/{multiprocessing.current_process().name}'
    try:
        dir_opers.makedir(params.instance.get('log_dir'))
        with open(config_filename) as f:
            config = json.load(f)
        for _, handler_config in config.get('handlers', {}).items():
            if 'filename' in handler_config:
                handler_config['filename'] = f'{params.instance.get("log_dir")}{process}/{handler_config["filename"]}'
        check_paths(config)
        logging.config.dictConfig(config)
        logger = logging.getLogger()
        logger.info('Logging.config is set from: ' + config_filename)
    except BaseException as ex:
        logger = get_default_logger()
        logger.exception(ex)


def get_default_logger():
    handler = RotatingFileHandler(
        filename=LOG_PATH, mode='a', maxBytes=10485760, backupCount=5)
    formatter = logging.Formatter('%(levelname)-7s %(asctime)s %(processName)-11s %(lineno)4d %(module)-15s : %(message)s')
    handler.setFormatter(formatter)
    res = logging.getLogger()
    res.setLevel('DEBUG')
    res.addHandler(handler)
    res.info('Default logger is created')
    return res


def check_paths(config):
    if params.instance.get('developing_mode') or is_pycharm():
        dir_opers.cleardir(params.instance.get('log_dir'))
    filenames = set([handler.get('filename') for handler in config.get('handlers').values()])
    filepaths = set([os.path.dirname(filename) for filename in filenames])
    for filepath in filepaths:
        if not os.path.exists(filepath):
            os.makedirs(filepath)



import asyncio
import logging
import signal

from peacepie import loglistener, msg_factory
from peacepie.assist import log_util
from peacepie.control import prime_admin, admin, safe_admin
from peacepie.control.inter import inter_server

INTER_COMMANDS = {'inter_connect', 'inter_disconnect'}


class HeadPrimeAdmin(prime_admin.PrimeAdmin):

    def __init__(self, parent, host_name, process_name):
        super().__init__(host_name, process_name)
        self.parent = parent
        self.is_head = True
        self.interlink = None
        self.safe_admin = None
        self.registered_handlers = {}

    async def pre_run(self):
        await super().pre_run()
        self.safe_admin = safe_admin.SafeAdmin(self)
        self.interlink = inter_server.InterServer(self)
        queue = asyncio.Queue()
        asyncio.get_running_loop().create_task(self.interlink.run(queue))
        await queue.get()
        loop = asyncio.get_running_loop()
        for signal_name in {'SIGINT', 'SIGTERM'}:
            loop.add_signal_handler(
                getattr(signal, signal_name),
                lambda: asyncio.create_task(self.adaptor.send(msg_factory.get_msg('quit')))
            )
        class_desc = {'requires_dist': 'peacepie.control.internal_starter'}
        body = {'class_desc': class_desc, 'name': 'internal_starter'}
        msg = self.adaptor.get_msg('create_actor', body, sender=self.adaptor.get_self_addr())
        await self.adaptor.send(msg)

    async def quit(self):
        if self.adaptor.stop_event is not None:
            return
        await self.process_admin.exit()
        self.adaptor.stop()

    async def exit(self):
        await self.actor_admin.exit()
        await self.intralink.exit()
        await self.interlink.exit()
        loglistener.instance.exit()

    async def handle(self, msg):
        command = msg.get('command')
        sender = msg.get('sender')
        if command == 'actor_is_created':
            body = msg.get('body')
            if not body:
                return False
            if body.get('entity') != 'internal_starter':
                return False
            await self.adaptor.send(self.adaptor.get_msg('start', recipient=body))
        elif command in INTER_COMMANDS:
            await self.interlink.queue.put(msg)
            logging.debug(log_util.async_sent_log(self, msg))
        elif command == 'get_credentials':
            await self.safe_admin.handle(msg)
        elif command == 'change_caches':
            await self.change_caches(msg)
        elif command == 'remove_from_caches':
            await self.remove_from_caches(msg)
        elif command == 'get_members':
            await self.get_members(msg)
        elif command == 'get_all_nodes':
            await self.get_all_nodes(sender, is_local=False)
        elif command == 'test_error':
            self.parent.set_test_error(msg.get('body'))
        else:
            return await super().handle(msg)
        return True

    async def change_caches(self, msg):
        recipient = msg.get('sender')
        body = msg.get('body') if msg.get('body') else {}
        await self.add_to_cache(body.get('node'), [body.get('entity')], True)
        links = [link for link in self.intralink.links]
        if links:
            await self.adaptor.group_ask(10, len(links),
                                         lambda index: {'command': 'change_cache', 'body': body,
                                                        'recipient': {'node': links[index], 'entity': None}}
            )
        if recipient:
            await self.adaptor.send(self.adaptor.get_msg('caches_are_changed', None, recipient))

    async def remove_from_caches(self, msg):
        recipient = msg.get('sender')
        body = msg.get('body') if msg.get('body') else {}
        await self.remove_from_cache(body, None)
        links = [link for link in self.intralink.links]
        await self.adaptor.group_ask(10, len(links),
                                     lambda index: {'command': 'remove_from_cache', 'body': body,
                                                    'recipient': {'node': links[index], 'entity': None}}
        )
        if recipient:
            await self.adaptor.send(self.adaptor.get_msg('removed_from_caches', None, recipient))

    async def get_members(self, msg):
        body = msg.get('body')
        page_size = body.get('page_size')
        level = body.get('level') if body.get('level') else 'head'
        xid = body.get('id') if body.get('id') else ''
        page = int(xid.split('_')[2]) if xid.startswith('_page_') else 0
        members = []
        if level == 'head':
            members = [{'next_level': 'prime', 'recipient': self.adaptor.name, 'id': self.adaptor.name}]
        if level == 'prime':
            members = self.intralink.get_members()
            members = [{'next_level': 'process', 'recipient': member, 'id': member} for member in members]
        elif level == 'process':
            members = self.process_admin.get_members()
            members = [{'next_level': 'actors', 'recipient': member, 'id': member} for member in members]
        elif level == 'actors':
            members = self.actor_admin.get_members()
            members = [{'next_level': 'actor', 'recipient': self.adaptor.name, 'id': member} for member in members]
        elif level == 'actor':
            members = [{'next_level': None, 'recipient': None, 'id': body.get('id')}]
        body = admin.format_members(level, self.adaptor.name, page_size, page, members)
        if level != 'head':
            body['_back'] = {'next_level': admin.get_prev(level), 'recipient': self.adaptor.name, 'id': '_back'}
        body['level'] = level
        ans = self.adaptor.get_msg('members', body, msg.get('sender'))
        await self.adaptor.send(ans)

    async def get_all_nodes(self, recipient, is_local=True):
        res = self.intralink.get_all_nodes()
        if is_local:
            return res
        else:
            await self.adaptor.send(self.adaptor.get_msg('all_nodes', res, recipient))
        return None



import inspect
import sys
from datetime import datetime


def is_testing():
    return "pytest" in sys.modules or "unittest" in sys.modules


def is_pycharm():
    return 'pycharm' in sys.executable.lower()


def get_current_time_string():
    now = datetime.now()
    return now.strftime('%H:%M:%S.%f')


def get_primary_class(module):
    classes = inspect.getmembers(module, inspect.isclass)
    if classes:
        primary_class_name, primary_class = classes[0]
        return primary_class
    else:
        return None



import logging

import json as default_json


local_json = default_json


def init(json_package):
    global local_json
    if json_package:
        local_json = json_package


def json_loads(jsn):
    if not jsn:
        return None
    res = None
    try:
        res = local_json.loads(jsn)
    except Exception as ex:
        logging.exception(ex, )
        logging.warning(f'The original string is "{jsn}"')
    return res


def json_dumps(jsn):
    if not jsn:
        return None
    res = None
    try:
        res = local_json.dumps(jsn)
    except Exception as ex:
        logging.exception(ex)
    return res



from peacepie.assist.serialization import Serializer


class InterQueue:

    def __init__(self, writer):
        self.writer = writer

    async def put(self, msg):
        self.writer.write(Serializer.serialize(msg))
        await self.writer.drain()

    async def close(self):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()
        self.writer = None



import asyncio
import logging

from peacepie import msg_factory
from peacepie.assist import log_util



class ActorAgent:

    def __init__(self, parent):
        self.parent = parent
        self.grandparent = parent.parent
        self.queue = asyncio.Queue()
        self.queue_storage = {}
        self.not_log_commands = set()
        self.cumulative_commands = {}
        logging.info(log_util.get_alias(self) + ' is created')

    async def run(self):
        while True:
            msg = await self.queue.get()
            logging.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    logging.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                logging.exception(ex)

    async def handle(self, msg):
        command = msg.get('command')
        if command == 'set_replica_params':
            await self.set_replica_params(msg)
        elif command == 'transit_message':
            await self.transit_message(msg)
        elif command == 'replica_resume':
            await self.replica_resume(msg)
        else:
            return False
        return True

    async def set_replica_params(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        entity = body.get('entity')
        params = body.get('params')
        recipient = msg.get('sender')
        actor = self.parent.actors.get(entity)
        adaptor = actor.get('adaptor')
        adaptor.set_params(params)
        if recipient:
            await self.grandparent.adaptor.send(msg_factory.get_msg('replica_params_are_set', None, recipient))

    async def transit_message(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        entity = body.get('entity')
        message = body.get('message')
        storage = self.queue_storage.get(entity)
        if not storage:
            storage = []
            self.queue_storage[entity] = storage
        storage.append(message)

    async def replica_resume(self, msg):
        body = msg.get('body') if msg.get('body') else {}
        entity = body.get('entity')
        recipient = msg.get('sender')
        storage = self.queue_storage.get(entity)
        if not storage:
            storage = []
            self.queue_storage[entity] = storage
        flag = True
        try:
            await self._replica_resume(entity, storage)
        except Exception as e:
            flag = False
            logging.exception(e)
        del self.queue_storage[entity]
        if recipient:
            if flag:
                await self.grandparent.adaptor.send(msg_factory.get_msg('replica_is_resumed', None, recipient))
            else:
                await self.grandparent.adaptor.send(msg_factory.get_msg('replica_is_not_resumed', None, recipient))

    async def _replica_resume(self, entity, storage):
        adaptor = self.parent.actors.get(entity).get('adaptor')
        try:
            while True:
                message = adaptor.queue.get_nowait()
                storage.append(message)
        except asyncio.QueueEmpty:
            pass
        for message in storage:
            adaptor.queue.put_nowait(message)
        adaptor.resume()



import asyncio
import threading


def thread_function(sync_function, sync_args, loop, result_future):
    try:
        result = sync_function(sync_args)
    except Exception as e:
        loop.call_soon_threadsafe(result_future.set_exception, e)
    else:
        loop.call_soon_threadsafe(result_future.set_result, result)


async def sync_as_async(sync_function, sync_args):
    loop = asyncio.get_event_loop()
    result_future = loop.create_future()
    thread = threading.Thread(target=thread_function, args=(sync_function, sync_args, loop, result_future))
    thread.start()
    result = await result_future
    return result



import asyncio
import logging
import os
import subprocess
import sys
from functools import partial

from peacepie import msg_factory, params, loglistener
from peacepie.assist import dir_opers, json_util, log_util, version


class PackageLoader:

    def __init__(self, parent):
        self.parent = parent
        self.not_log_commands = set()
        self.cumulative_commands = {}
        self.queue = asyncio.Queue()
        self.source_path = params.instance.get('source_path')
        dir_opers.makedir(self.source_path)
        self.tmp_path = f'{params.instance["package_dir"]}/tmp'
        dir_opers.makedir(self.tmp_path, clear=True)
        self.loadings = {}
        logging.info(log_util.get_alias(self) + ' is created')

    async def run(self, queue):
        await queue.put(msg_factory.get_msg('ready'))
        while True:
            msg = await self.queue.get()
            logging.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    logging.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                logging.exception(ex)

    async def handle(self, msg):
        command = msg['command']
        if command == 'load_package':
            await self.load_package(msg)
        else:
            return False
        return True

    async def load_package(self, msg):
        recipient = msg.get('sender')
        timeout = msg.get('timeout')
        body = msg.get('body')
        requires_dist = body.get('requires_dist')
        bundle = await self.acquire_package(requires_dist, body.get('url'), timeout)
        if bundle:
            self.move_packages()
            body = {'bundle': list(bundle)}
            await self.parent.adaptor.send(msg_factory.get_msg('package_is_loaded', body, recipient), self)
        else:
            await self.parent.adaptor.send(msg_factory.get_msg('package_is_not_loaded', None, recipient), self)

    def move_packages(self):
        for package in os.listdir(self.tmp_path):
            dir_opers.move_dir(f'{self.tmp_path}/{package}', f'{self.source_path}/{package}')

    async def acquire_package(self, requires_dist, url, timeout):
        package_name = requires_dist.get('package_name')
        version_spec = requires_dist.get('version_spec')
        package_loading = self.loadings.get(package_name)
        if package_loading is None:
            bundle = self.find_bundle(package_name, version_spec)
            if bundle:
                return bundle
        else:
            queue = asyncio.Queue()
            package_loading.append(queue)
            await queue.get()
            return self.find_bundle(package_name, version_spec)
        self.loadings[package_name] = []
        await self.install_package(package_name, version_spec, url, timeout)
        ver, dependencies = self.find_version_with_dependencies(package_name, version_spec)
        if not ver:
            return None
        bundle = {f'{package_name}-{ver}'}
        tasks = [asyncio.create_task(self.acquire_package(dependency, url, timeout))
                 for dependency in dependencies]
        if len(tasks) > 0:
            results = await asyncio.gather(*tasks)
            for result in results:
                if result is None:
                    return None
                bundle = bundle.union(result)
        self.save_bundle(package_name, ver, bundle)
        for queue in self.loadings.get(package_name):
            await queue.put(None)
        del self.loadings[package_name]
        return bundle

    def find_bundle(self, package_name, version_spec):
        res = self._find_bundle(self.source_path, package_name, version_spec)
        if res:
            return res
        return self._find_bundle(self.tmp_path, package_name, version_spec)

    def _find_bundle(self, path, package_name, version_spec):
        ver = None
        bundle = None
        for pack in os.listdir(path):
            pack_path = f'{path}/{pack}'
            for entry in os.listdir(pack_path):
                entry_path = f'{pack_path}/{entry}'
                if not (os.path.isdir(entry_path) and entry.endswith('.dist-info')):
                    continue
                name, v, _ = dir_opers.get_metadata(entry_path)
                if name.lower().replace('-', '_') == package_name.lower().replace('-', '_'):
                    if version.check_version(v, version_spec):
                        if ver is None or version.check_version(v, f'>{ver}'):
                            ver = v
                            bundle = get_bundle(entry_path)
        return bundle

    def find_version_with_dependencies(self, package_name, version_spec):
        ver = None
        dependencies = None
        for pack in os.listdir(self.tmp_path):
            pack_path = f'{self.tmp_path}/{pack}'
            for entry in os.listdir(pack_path):
                entry_path = f'{pack_path}/{entry}'
                if not (os.path.isdir(entry_path) and entry.endswith('.dist-info')):
                    continue
                name, v, ds = dir_opers.get_metadata(entry_path)
                if name.lower().replace('-', '_') == package_name.lower().replace('-', '_'):
                    if version.check_version(v, version_spec):
                        if ver is None or version.check_version(v, f'>{ver}'):
                            ver = v
                            dependencies = ds
                    if not pack_path.endswith(f'-{v}'):
                        dir_opers.rename_dir(pack_path, f'{pack_path}-{v}')
        return ver, dependencies

    async def install_package(self, package_name, version_spec, url, timeout):
        ver_package_name = package_name + version_spec if version_spec else package_name
        args = ([sys.executable, '-m', 'pip', 'install', '--no-deps', '--disable-pip-version-check'])
        if url:
            host = url.split("//")[-1].split("/")[0]
            args.append(f'--trusted-host={host}')
            args.append(f'--extra-index-url={url}')
        args.append(f'--target={self.tmp_path}/{package_name}')
        args.append(ver_package_name)
        logging.info(args)
        loop = asyncio.get_running_loop()
        try:
            output = await asyncio.wait_for(
                loop.run_in_executor(None,
                                     partial(subprocess.check_output,args, stderr=subprocess.STDOUT, text=True)),
                timeout=timeout
            )
            logging.info(self.parent.adaptor.res_squeeze(output))
        except Exception as e:
            logging.exception(e)

    def save_bundle(self, package_name, package_ver, bundle):
        for package in os.listdir(self.tmp_path):
            package_path = f'{self.tmp_path}/{package}'
            for entry in os.listdir(package_path):
                entry_path = f'{package_path}/{entry}'
                if not (os.path.isdir(entry_path) and entry.endswith('.dist-info')):
                    continue
                name, ver, _ = dir_opers.get_metadata(entry_path)
                if name.lower().replace('-', '_') == package_name.lower().replace('-', '_') and ver == package_ver:
                    lines = list(bundle)
                    lines.sort()
                    with open(f'{entry_path}/BUNDLE', 'w', encoding='utf-8') as f:
                        f.writelines(line + '\n' for line in lines)
                    logging.info(f'BUNDLE is saved to {entry_path}')
                    return


def get_bundle(path):
    try:
        with open(f'{path}/BUNDLE', 'r', encoding='utf-8') as f:
            bundle = [line.strip() for line in f.readlines()]
        return bundle
    except Exception as e:
        logging.exception(e)



import asyncio
import logging
import importlib
import os
import time

from peacepie import params, msg_factory
from peacepie.assist import auxiliaries, dir_opers, log_util, version

shared_folders = {'__pycache__', 'bin'}


class PackageAdmin:

    def __init__(self, parent):
        self.parent = parent
        self.not_log_commands = set()
        self.cumulative_commands = {}
        self.waiters = {}
        self.work_path = f'{params.instance["package_dir"]}/work/{self.parent.parent.process_name}'
        dir_opers.adjust_path(params.instance['package_dir'], self.parent.parent.process_name, shared_folders)
        logging.info(log_util.get_alias(self) + ' is created')

    async def get_class(self, class_desc, timeout):
        if not timeout:
            timeout = 120
        try:
            pack = await self.get_package(class_desc, timeout)
            if not pack:
                return None
            if class_desc.get('class'):
                return getattr(pack, class_desc.get('class'))
            else:
                return auxiliaries.get_primary_class(pack)
        except Exception as ex:
            logging.exception(ex)
        return None

    async def get_package(self, class_desc, timeout):
        requires_dist = version.parse_requires_dist(class_desc.get('requires_dist'))
        package_name = requires_dist.get('package_name')
        version_spec = requires_dist.get('version_spec')
        if not class_desc.get('class'):
            return importlib.import_module(package_name)
        is_exist = self.check_version(package_name, version_spec)
        if is_exist is not None:
            if is_exist:
                return importlib.import_module(package_name)
            else:
                return None
        if params.instance.get('developing_mode'):
            res = await self.developing_symlink(package_name, version_spec)
            if res:
                return res
        res = await self.retrieve_package(requires_dist, class_desc.get('extra-index-url'), timeout)
        return res

    def check_version(self, package_name, version_spec):
        for entry in os.listdir(self.work_path):
            entry_path = f'{self.work_path}/{entry}'
            if not (os.path.isdir(entry_path) and entry.endswith('.dist-info')):
                continue
            name, ver, _ = dir_opers.get_metadata(entry_path)
            if name == package_name:
                return version.check_version(ver, version_spec)
        return None

    async def developing_symlink(self, package_name, version_spec):
        path = f'{params.instance.get("plugin_dir")}/{package_name}'
        vers = [name for name in os.listdir(path) if version.version_from_string(name)]
        ver = version.find_max_version(vers, version_spec)
        src = f'{path}/{ver}'
        if dir_opers.is_dir_exist(f'{src}/src'):
            src = f'{src}/src/{package_name}'
        else:
            src = f'{src}/{package_name}'
        if await dir_opers.create_symlink(src, f'{self.work_path}/{package_name}'):
            self.create_package_info(package_name, ver)
            return importlib.import_module(package_name)
        return None

    def create_package_info(self, package_name, ver):
        package_info_path = f'{self.work_path}/{package_name}-{ver}.dist-info'
        dir_opers.recreatedir(package_info_path)
        with open(f'{package_info_path}/METADATA', 'w', encoding='utf-8') as f:
                 f.write(f'Name: {package_name}\n')
                 f.write(f'Version: {ver}\n')

    async def retrieve_package(self, requires_dist, url, timeout):
        queue = asyncio.Queue()
        if not url:
            url = params.instance.get('extra-index-url')
        waiter = {'requires_dist': requires_dist, 'url':url, 'timeout': timeout, 'queue': queue}
        package_name = requires_dist.get('package_name')
        waiters = self.waiters.get(package_name)
        if waiters:
            active = waiters.pop()
            waiters.append(waiter)
            waiters.sort(key=lambda x: x['timeout'])
            waiters.append(active)
        else:
            waiters = [waiter]
            self.waiters[package_name] = waiters
            asyncio.get_running_loop().create_task(self.acquire_package(package_name, waiters))
        res = await queue.get()
        return res

    async def acquire_package(self, package_name, waiters):
        start = time.time()
        while waiters:
            active = waiters[-1]
            if active.get('timeout') < time.time() - start:
                for waiter in waiters:
                    await waiter.get('queue').put(None)
                break
            res = await self.load_package(active)
            if res:
                for waiter in waiters:
                    await waiter.get('queue').put(res)
                break
            else:
                await active.get('queue').put(None)
            waiters.pop()
        del self.waiters[package_name]

    async def load_package(self, active):
        recipient = self.parent.parent.adaptor.get_head_addr()
        requires_dist = active.get('requires_dist')
        package_name = requires_dist.get('package_name')
        version_spec = requires_dist.get('version_spec')
        body = {'requires_dist': requires_dist, 'url': active.get('url')}
        msg = msg_factory.get_msg('load_package', body, recipient)
        ans = await self.parent.parent.adaptor.ask(msg, active.get('timeout'), self)
        if ans.get('command') == 'package_is_loaded':
            try:
                bundle = ans.get('body').get('bundle')
                packages = {}
                for member in bundle:
                    name, ver = member.rsplit('-', 1)
                    packages[name] = ver
                need_to_link = self.need_to_link(package_name, version_spec, packages)
                if await self.link_packages(need_to_link):
                        return importlib.import_module(package_name)
            except Exception as e:
                logging.exception(e)
        return None

    def need_to_link(self, package_name, version_spec, packages):
        is_exist = self.check_version(package_name, version_spec)
        if is_exist is not None:
            if is_exist:
                return {}
            else:
                return None
        ver = packages.get(package_name)
        if not ver:
            return None
        result = {package_name: ver}
        dependencies = self.get_dependencies(package_name, ver)
        for dependency in dependencies:
            res = self.need_to_link(dependency.get('package_name'), dependency.get('version_spec'), packages)
            if res is None:
                return None
            else:
                result.update(res)
        return result

    def get_dependencies(self, package_name, package_ver):
        path = f'{params.instance.get("source_path")}/{package_name}-{package_ver}'
        for entry in os.listdir(path):
            entry_path = f'{path}/{entry}'
            if not (os.path.isdir(entry_path) and entry.endswith('.dist-info')):
                continue
            _, _, dependencies = dir_opers.get_metadata(entry_path)
            return dependencies
        return []

    async def link_packages(self, packages):
        if packages is None:
            return False
        for package_name in packages:
            src_path = f'{params.instance.get("source_path")}/{package_name}-{packages.get(package_name)}'
            dst_path = f'{self.work_path}'
            await dir_opers.link_package(f'{src_path}', f'{dst_path}', shared_folders)
        return True




import asyncio
import logging
import os
import subprocess
import sys

from peacepie import params, msg_factory
from peacepie.assist import log_util, dir_opers, version

DELIVERY_DIR = 'delivery'

BLOCK_SIZE = 1024


class Delivery:

    def __init__(self, parent):
        self.parent = parent
        self.delivery_path = f'{params.instance["package_dir"]}/{DELIVERY_DIR}'
        dir_opers.makedir(self.delivery_path, True)
        self.queue = asyncio.Queue()
        logging.info(log_util.get_alias(self) + ' is created')

    async def run(self):
        while True:
            msg = await self.queue.get()
            logging.debug(log_util.async_received_log(self, msg))
            try:
                if not await self.handle(msg):
                    logging.warning(log_util.get_alias(self) + ': The message is not handled: ' + str(msg))
            except Exception as ex:
                logging.exception(ex)

    async def handle(self, msg):
        command = msg.get('command')
        if command == 'deliver_package':
            await self.deliver_package(msg)
        elif command == 'transfer':
            await self.transfer(msg)
        else:
            return False
        return True

    async def deliver_package(self, msg):
        dir_opers.cleardir(self.delivery_path)
        body = msg.get('body')
        package_desc = body.get('package_desc')
        if (not await self.download(package_desc.get('package_name'), package_desc.get(version.VERSION)) or
                not await self.send(body.get('recipient'))):
            ans = self.parent.adaptor.get_msg('package_is_not_delivered', recipient=msg.get('sender'))
            await self.parent.adaptor.send(ans)
            dir_opers.cleardir(self.delivery_path)
            return
        ans = self.parent.adaptor.get_msg('package_delivered', recipient=msg.get('sender'))
        await self.parent.adaptor.send(ans)
        dir_opers.cleardir(self.delivery_path)

    async def download(self, package_name, conditions):
        package_name = package_name + version.conditions_as_text(conditions)
        if self._download(package_name):
            return True
        recipient = self.parent.adaptor.get_head_addr()
        msg = msg_factory.get_msg('load_package', {'requires_dist': package_name}, recipient=recipient)
        ans = await self.parent.adaptor.ask(self, msg)
        if ans['command'] == 'package_is_not_loaded':
            return False
        return self._download(package_name)

    def _download(self, package_name):
        res = None
        try:
            res = subprocess.check_call(
                [sys.executable, '-m', 'pip', 'download', package_name, '--disable-pip-version-check', '--no-index',
                 f'--find-links={params.instance["package_dir"]}/source', f'-d{self.delivery_path}'])
        except Exception as ex:
            logging.exception(ex)
        return res == 0

    async def send(self, recipient):
        for filename in os.listdir(self.delivery_path):
            with open(f'{self.delivery_path}/{filename}', 'br') as f:
                while True:
                    data = f.read(BLOCK_SIZE)
                    if not data:
                        break
                    body = {'filename': filename, 'data': data.decode('latin-1')}
                    msg = msg_factory.get_msg('transfer', body, recipient=recipient)
                    ans = await self.parent.adaptor.ask(self, msg, 10)
                    if ans.get('command') != 'transferred':
                        return False
        return True

    async def transfer(self, msg):
        body = msg.get('body')
        with open(f'{self.delivery_path}/{body.get("filename")}', 'ba') as f:
            f.write(body.get('data').encode('latin-1'))
        await self.parent.adaptor.send(self, msg_factory.get_msg('transferred', recipient=msg.get('sender')))





import logging
import multiprocessing

from peacepie.assist import log_util

instance = None


def init_multimanager(name):
    global instance
    instance = MultiManager(name)


class MultiManager:

    def __init__(self, name):
        self.logger = logging.getLogger()
        self.name = name
        self.manager = multiprocessing.Manager()
        self.logger.info(f'{log_util.get_alias(self)} is created')

    def get_queue(self):
        return self.manager.Queue()



import asyncio
import logging

from peacepie import adaptor, params, multimanager, loglistener, msg_factory, log_conf
from peacepie.assist import log_util, json_util
from peacepie.control import head_prime_admin, prime_admin


class PeaceSystem:

    def __init__(self, path, test_params=None, json_package=None):
        json_util.init(json_package)
        params.init_params(path, test_params)
        log_conf.logger_start(params.instance.get('log_config'))
        self.is_head = params.instance.get('intra_role') == 'master'
        self.host_name = params.instance.get('host_name')
        self.process_name = params.instance.get('process_name')
        self.prefix = self.host_name + '.' + self.process_name + '.'
        multimanager.init_multimanager(self.prefix + 'multimanager')
        loglistener.init_loglistener(self.prefix + 'loglistener')
        self.name = self.prefix + 'system'
        msg_queue = multimanager.instance.get_queue()
        msg_queue.put(0)
        self.task = None
        self.test_errors = []
        msg_factory.init_msg_factory(self.host_name, self.process_name, 'msg_factory', msg_queue)
        logging.info(log_util.get_alias(self) + ' is created')

    async def start(self):
        if self.is_head:
            performer = head_prime_admin.HeadPrimeAdmin(self, self.host_name, self.process_name)
        else:
            performer = prime_admin.PrimeAdmin(self.host_name, self.process_name)
        actr = adaptor.Adaptor(None, self.prefix + 'admin', None, performer)
        self.task = asyncio.get_running_loop().create_task(actr.run())

    def set_test_error(self, res):
        self.test_errors.append(res)
